{*******************************************
 *  Rem's Project B - for C3 hardware v020 *
 *******************************************

Update history:
- First 'test demo movie player' release: november 2010
- Public release of unfinished game engine: january 24th, 2012

GFX engine overhaul:
- Tilemap system:
        Each vertical line now have a 'vertical offset' indicating which tile line must be displayed.
        This flexible system allows for arbitrary vertical scaling, repeating lines, variable configuration of splitscreen(s),
        'mirror effect', etc.
        Also a 'horizontal offset' allowing per scanline horizontal scrolling of tilemap.
        Can be used to make wave effect, shearing, multi-level parallax scrolling, etc.  

' Before 'WhichSprite' optimisation: Free:91 longs
' After 'WhichSprite' in byte: Free 71 longs! pathetic: byte variables consume WAY more ram due to syntax cast and shit
' After 'WhichSprite' in long: Free 95 longs. Only 4 longs = 16 bytes more. 
}

CON

  _clkmode = xtal1 + pll16x  ' Set clock multiplier to 16x
  _xinfreq = 5_000_000 ' Set to 5Mhz
  _memstart = $10                   ' memory starts $10 in!
  _STACK = 74

  ' Warning: In order to add a variable here in 'gfx_buffer', you need to ensure that there is enough
  ' cog RAM allocated. See 'padding' in REM_drv_db_c3_fsrmfemto.
  spin_cog = 0
  spin_request_scanline = 1
  spin_vertical_scroll = 2
  spin_horizontal_scroll = 3
  spin_sprites_adr = 4
  spin_tilemap_adr = 5
  spin_tiles_adr = 6
  spin_gamepad = 7
  spin_tv_communication = 8
  spin_display_base = 9

  ' Gamepad constant declaration
  NES_RIGHT  = %00000001
  NES_LEFT   = %00000010
  NES_DOWN   = %00000100
  NES_UP     = %00001000
  NES_START  = %00010000
  NES_SELECT = %00100000
  NES_B      = %01000000
  NES_A      = %10000000

  ' Gameplay constants
  CtrlLeft      = NES_LEFT'1
  CtrlRight     = NES_RIGHT'2
  CtrlUp        = NES_UP'4
  CtrlDown      = NES_DOWN'8
  CtrlJump      = NES_B'16
  CtrlShoot     = NES_A'32
  FLOOROFFSET   = 7
  SPRITE_BULLET_FIRST=12
  SPRITE_BULLET_LAST=23
  SPRITE_BULLET_NB=(SPRITE_BULLET_LAST-SPRITE_BULLET_FIRST+1)
  
  'Exported CONSTANT address for WAD file stored on SD:
  stream_map = 0 ' File stream.map
  stream_map_size = 4096
  tiles_dat = 4096 ' File tiles.dat
  tiles_dat_size = 12288
  anim_tiles_dat = 16384 ' File anim_tiles.dat
  anim_tiles_dat_size = 16384
  bot_sprite_dat = 32768 ' File bot_sprite.dat
  bot_sprite_dat_size = 55296
  ruffian_sprite_dat = 88064 ' File ruffian_sprite.dat
  ruffian_sprite_dat_size = 55296
  sprite16x16_dat = 143360 ' File sprite16x16.dat
  sprite16x16_dat_size = 3072
  sprite8x8_dat = 146432 ' File sprite8x8.dat
  sprite8x8_dat_size = 1024
  BGscrollSpeed_dat = 147456 ' File BGscrollSpeed.dat
  BGscrollSpeed_dat_size = 192
  remmap_map = 147648 ' File remmap.map
  remmap_map_size = 1024
  moviecol_pal = 148672 ' File moviecol.pal
  moviecol_pal_size = 102
  movie_cut = 148774 ' File movie.cut
  movie_cut_size = 1154002
  gameover_cut = 1302776 ' File gameover.cut
  gameover_cut_size = 153768
  HDMF_song_bin = 1456544 ' File HDMF_song.bin
  HDMF_song_bin_size = 732
  'End of file entries. Total size of WAD file=1457276

  ' Pointer into SRAM memory (65536 bytes):
  SRAM_music_addr = 0
  SRAM_Stream_map_addr = 8192
  SRAM_GFX_tile_mode_addr = 16384
  SRAM_GFX_cutscene_mode_addr = 18432
  SRAM_Anim_tiles_addr = 20480
  SRAM_Movie_Table_addr = 36864

  ' Pointer into FLASH memory (1048576 bytes):
  FLASH_sprite16x16_addr = 0
  FLASH_sprite8x8_addr = 3072
  FLASH_player_sprite_addr = 4096
  FLASH_ruffian_sprite_addr = 59392 

VAR

  ' General purpose spin temporary variable
  long WhichSprite ' global optimisation when refering to 'current sprite being processed'
  long WhichSprite4 ' cache value for WhichSprite<<2

  long temp1
  long temp2
  long temp3
  word sprites[4*gfx#MAX_NB_SPRITE] ' Each sprite takes 8 bytes = 4 words, allocate space for 24 sprites
  ' sprite are organised like this in memory:
  ' word 16-bit[0]: sprite X position, fixed-point: x<<7
  ' word 16-bit[1]: sprite Y position, fixed-point: y<<7
  ' word 16-bit[2]: sprite size: (width<<8) + height
  ' word 16-bit[3]: sprite address
  long ioControl[5]   ' SPI
  long gfxbuffer      ' pointer to gfx buffer (which contains pointers to sprites, vertical scroll, etc.)
  long FrameCounter
  byte music_buffer[32]  ' 32 byte song buffer
  long music_ptr
  byte SpriteAnimDelay
  'long map_stream_ptr
  long anim_tiles_ptr
  long spritesFrame[11] ' immediate frame offset NOTE: on C3, sprite 12 to 23 don't have a 'spritesFrame' since they are clones
                        ' note: sprite 5 to 8 shouldn't have spritesFrame either: we are wasting 4*4 = 16 bytes of RAM here 
  long spritesPosX[gfx#MAX_NB_SPRITE] ' sprite game position x (7-bit fixed point)
  long spritesPosY[gfx#MAX_NB_SPRITE] ' sprite game position y (7-bit fixed point)
  byte spritesDirX[SPRITE_BULLET_NB] ' sprite speed/direction x (7-bit fixed point)
  byte spritesDirY[SPRITE_BULLET_NB] ' sprite speed/direction y (7-bit fixed point)
  byte spritesFlip[3] ' immediate sprite flip
  byte spritesHotX[3] ' sprite horizontal hotspot offset (used for off-centered sprites)

  byte cogid_gfx[gfx#GFX_COG_TOTAL] ' Array of id to know which cogs are used for gfx engine

  word movieLen
  long movie_frame
  long movie_seek_loop
  long movie_frame_cnt

  long JumpSpeed[3]
  byte JumpCounter[3]
  byte JumpRelease[3]
  byte currentanim[3]
  byte ShootCounter[3]
  byte AutoFireDelay[3]
  byte AnimUpdateImportant[3]
  byte AICounter[3]
  byte HitPoint[3]
  byte HMoveSpeed[3]
  byte JumpPowerStart[3]
  byte JumpPowerBoost[3]
  word GameTimer
  word GameTimerCache
  word GameTimerFrame
  word Score
  word ScoreCache
  word HiScore
  byte HitPointCache
  byte HitPointBlink
  byte CurrentHud
  byte CurrentHudCache
  byte HudSlide
  byte GameState
  long GameStateTimer
  byte ComputerHacked

  byte SpawnTiles[128] ' Array of spawn tile (should have been 256 but we are out of ram)

  byte Controls

  word MapWidth

  long SpinGamepad ' optimisation: spin_gamepad is no longer used in gfx_buffer
  word vscroll193
  word hscroll193
  word vscroll194
  word hscroll194

OBJ
  
  fsrw: "REM_drv_db_c3_fsrwfemto_020.spin"                      ' Modified SD routines
  'key   : "REM_drv_keyboard_020.spin"
  tv    : "REM_drv_tv_020.spin"
  gfx   : "REM_drv_gfx_engine_020.spin"
  cut   : "REM_drv_cutscene_engine_020.spin"
  snd : "REM_drv_EPM_HDMF_020.spin"      ' HDMF Lite driver
 
{ Cog allocation:
cog 0 = SPIN main code
cog 1 = SD/Flash/Sram driver
cog 2 = tv driver
cog 3..4..5..6 = GFX cogs
cog 7 = Sound driver
}
PUB start

  DIRA[15] := 1 ' Set debug LED ready for output         
  'outa[15] := 0 ' Use this to lit the debug led
  cogid_gfx[0] := 255 ' magic value to indicate gfx cog id are unspecified yet.
  
  ' Start SD/SPI driver early, so we can overwrite its HUB-ram code space with runtime buffer 
  fsrw.start(@ioControl)                                  ' Start SPI driver
  ' debug: flash erase can be omitted while developping since flash will preserve its content
  ' note that the call to 'LoadFlashData' can be skipped too
  fsrw.flash_chip_erase_Start   ' Begin flash erase immediately so we don't have to wait too much until its done
  
  gfxbuffer := fsrw.RV_getcog_overlap+1024               ' Put GFX buffer overlapping after spi driver buffers

  ' Copy gfx cog-code into SRAM:
  ' Copy 'tile mode'
  fsrw.writeSram(SRAM_GFX_tile_mode_addr, gfx.RV_getcog_overlap, 2048)
  ' Copy 'cutscene mode'
  fsrw.writeSram(SRAM_GFX_cutscene_mode_addr, cut.RV_getcog_overlap, 2048)

  snd.start(0)
   
  ' Boot requested number of rendering cogs:
  ' 'GFX_COG_TOTAL' values are:
  ' 1 = Unnacceptable: screen will flicker horrendously
  ' 2 = No sprite, i.e.: just enough cpu power to render the tilemap
  ' 3 = Couple of sprites.
  ' 4 = A lot of sprites.
  ' If we don't provide enough rendering cogs to draw the sprites, there will be some missing sprite horizontal lines
  ' and the debug LED will light up to indicate you need more rendering cogs, or less sprite on the same line, or
  ' horizontally smaller sprites.

  ' Rem GFX engine setup variables:
  long[gfxbuffer][spin_tilemap_adr] := @tilemap     ' Tilemap address
  long[gfxbuffer][spin_tiles_adr] := @tiles         ' Tiles address
  long[gfxbuffer][spin_sprites_adr] := @sprites         ' Sprites address

  ' Both scroll array are stored as 195 words (390 bytes):
  ' 192 'word': one per scanline, defining an offset
  ' 1 additional 'word': base offset (this is added with 'per scanline' to give final offset), for background map
  ' 1 additional 'word': foreground offset (for foreground map)
  ' 1 additional 'word': start and stop scanline for foreground map
  long[gfxbuffer][spin_vertical_scroll] := gfx.RV_getcog_overlap+1024  ' Pointer to free HUB memory to store vertical scroll array
  long[gfxbuffer][spin_horizontal_scroll] := snd.RV_getcog_overlap+1280 ' Pointer to free HUB memory to store horizontal scroll array
  
  InitSpritePointers

  ' Start keyboard driver OR mouse driver, but not both at the same time
  'key.start(26, 27)
  
  word[long[gfxbuffer][spin_vertical_scroll]][192] := 0            ' BG vertical base offset
  word[long[gfxbuffer][spin_horizontal_scroll]][192] := 0          ' BG horizontal base offset
  word[long[gfxbuffer][spin_vertical_scroll]][193] := 0            ' foreground vertical base offset
  word[long[gfxbuffer][spin_horizontal_scroll]][193] := 0          ' foreground horizontal base offset
  'word[long[gfxbuffer][spin_vertical_scroll]][194] := 0            ' foreground start scanline
  'word[long[gfxbuffer][spin_horizontal_scroll]][194] := 191          ' foreground stop scanline
  ResetBGScroll
  
  ' Start tv driver
  tv_gfxbuffer := gfxbuffer
  tv.start(@tvparams)

  'tv_color := $0C  'blue
  ' start gfx engine in 'tile mode'
  SwitchGfxEngine(0)

  vscroll194 := 191            ' foreground start scanline
  hscroll194 := 191          ' foreground stop scanline

  hex_display_word(CONSTANT(24+3*gfx#MAPSIZEX), fsrw.RV_getcog_end_of_ram)
  hex_display_word(CONSTANT(24+4*gfx#MAPSIZEX), gfx.RV_getcog_end_of_ram)
  hex_display_word(CONSTANT(24+5*gfx#MAPSIZEX), snd.RV_getcog_end_of_ram)
  hex_display_word(CONSTANT(24+6*gfx#MAPSIZEX), tv.RV_getcog_end_of_ram)
  hex_display_word(CONSTANT(24+7*gfx#MAPSIZEX), cut.RV_getcog_end_of_ram)

  'word[long[gfxbuffer][spin_horizontal_scroll]][194] := 16
  MountSD
  'RefreshFromSD
  
  CutsceneMode(movie_cut)
  word[long[gfxbuffer][spin_horizontal_scroll]][194] := 0 ' enable display, but keep foreground hidden
  
  ' Wait until flash erase finishes.
  DisplayString(CONSTANT(@tilemap+_memstart + 8+0*gfx#MAPSIZEX), string("FLASH>ERASE>>>"))
  fsrw.flash_chip_erase_WaitDone
  ' loadflashdata can be skipped too when developping since the flash memory doesn't get erased between shutdown, etc.
  ' However, if you load another game (like Ranquest, etc.) that uses flash, this MUST be enabled.
  LoadFlashData

  PrepareVar
  DisplayString(CONSTANT(@tilemap+_memstart + 8+0*gfx#MAPSIZEX), string(">>>>>>>>>>>>>>>"))
  
  ' Start of main spin loop here.
  repeat
    SpinGamepad := NES_Read_Gamepad
    SpriteAnimDelay++

    wait_vbl
    
    processGameState

    word[long[gfxbuffer][spin_vertical_scroll]][193] := vscroll193            ' foreground vertical base offset
    word[long[gfxbuffer][spin_horizontal_scroll]][193] := hscroll193          ' foreground horizontal base offset
    word[long[gfxbuffer][spin_vertical_scroll]][194] := vscroll194            ' foreground start scanline
    word[long[gfxbuffer][spin_horizontal_scroll]][194] := hscroll194          ' foreground stop scanline

    updateHUD
    animatetiles
    rendersprites
    
    FrameCounter++
    If SpriteAnimDelay == 6
      SpriteAnimDelay := 0

  ' Repeat forever

    {
    ' Read both gamepad
    temp1 := NES_Read_Gamepad
    ' Arrow pad is mapped over gamepad #1 controls
    {
    if(key.keystate($75))
      temp1|=CONSTANT(NES_UP)
    elseif(key.keystate($72))          
      temp1|=CONSTANT(NES_DOWN)
    if(key.keystate($6B))
      temp1|=CONSTANT(NES_LEFT)
    elseif(key.keystate($74))
      temp1|=CONSTANT(NES_RIGHT)}

    if temp1 & NES_LEFT
      sprites[CONSTANT(0 + 0*4)] -= 128
      sprite_dir := 1
    elseif temp1 & NES_RIGHT
      sprites[CONSTANT(0 + 0*4)] += 128
      sprite_dir := 0
      
    if temp1 & NES_UP
      sprites[CONSTANT(1 + 0*4)] -= 128
    elseif temp1 & NES_DOWN
      sprites[CONSTANT(1 + 0*4)] += 128

    if temp1&NES_A and not (long[gfxbuffer][spin_gamepad]&NES_A)
      snd.SFX_start(snd#SFX_STATE__JUMP)

    if temp1&NES_SELECT and not (long[gfxbuffer][spin_gamepad]&NES_SELECT)
      CutsceneMode

    if temp1&NES_START and not (long[gfxbuffer][spin_gamepad]&NES_START)
      RefreshFromSD
      
    long[gfxbuffer][spin_gamepad] := temp1
    
    temp2 := (sprites[CONSTANT(0 + 0*4)] >> 7) - 100
    'tv_color := temp2
    hex_display_nibble(1, temp2&$0F)   ' extract low digit     

    temp2 := (sprites[CONSTANT(1 + 0*4)] >> 7) - 100
    'tv_color += temp2<<4
    hex_display_nibble(0, temp2&$0F)   ' extract high digit     

    ' Enable this to debug pressed keyboard scancode on the screen
    {repeat temp1 from 0 to 254
      if key.keystate(temp1)
        hex_display_byte(CONSTANT(4+0), temp1)  }

    hex_display_long(CONSTANT(2+1*gfx#MAPSIZEX), FrameCounter)
    hex_display_long(CONSTANT(14+1*gfx#MAPSIZEX), fsrw.flashSize)
    hex_display_word(CONSTANT(24+1*gfx#MAPSIZEX), word[long[gfxbuffer][spin_horizontal_scroll]][193])
    
    'tv_color := $4B

    if(long[gfxbuffer][spin_gamepad]&NES_A)
      'tv_color := $8C           ' yellowish: time 'sram' loading time
      'fsrw.readSram(CONSTANT(194*2), long[gfxbuffer][spin_horizontal_scroll], CONSTANT(194*2))
      'tv_color := $9C           ' yellowish: time 'sram' loading time
      'fsrw.reamSRAMFast0(CONSTANT(194*2), long[gfxbuffer][spin_horizontal_scroll], CONSTANT(194*2))
      'tv_color := $AC           ' yellowish: time 'sram' loading time
      'fsrw.readData(CONSTANT(194*2), long[gfxbuffer][spin_horizontal_scroll], CONSTANT(194*2))
      
      tv_color := $05   
    elseif(long[gfxbuffer][spin_gamepad]&NES_B) 
      ' Read 192 block of 512 bytes (total: 98304 bytes) then seek back to start of file
      'tv_color := $8C           ' yellowish: time seek time
 '     repeat temp4 from 0 to 3  ' loop reading the whole file 4 times
'        repeat temp3 from 0 to 127
      'temp3 := (sprites.word[CONSTANT(1 + 0*4)] >> 7) - 100
      'fsrw.seek(stream_map + temp3*32)
      
    else
      ' test scroll screen with gamepad
      {repeat temp1 from 0 to 90
        'spin_y_offsets[temp1] := temp1
        word[long[gfxbuffer][spin_vertical_scroll]][temp1] := (sprites.word[CONSTANT(1 + 0*4)] >> 7) - 100 + temp1
        word[long[gfxbuffer][spin_horizontal_scroll]][temp1] := (sprites.word[CONSTANT(0 + 0*4)] >> 7) - 100}
      'Scroll foreground map:        
      'word[long[gfxbuffer][spin_vertical_scroll]][193] := (sprites.word[CONSTANT(1 + 0*4)] >> 7) - 100 #> 0 <# 80 
      'word[long[gfxbuffer][spin_horizontal_scroll]][193] := (sprites.word[CONSTANT(0 + 0*4)] >> 7) - 100

      'Scroll background map slowly
      'word[long[gfxbuffer][spin_vertical_scroll]][192] := ((sprites.word[CONSTANT(1 + 0*4)] >> 7) - 100) >> 3
      if(long[gfxbuffer][spin_gamepad]&NES_UP)
        word[long[gfxbuffer][spin_vertical_scroll]][193] := word[long[gfxbuffer][spin_vertical_scroll]][193]-1 #> 0
      elseif(long[gfxbuffer][spin_gamepad]&NES_DOWN)
        word[long[gfxbuffer][spin_vertical_scroll]][193] := word[long[gfxbuffer][spin_vertical_scroll]][193]+1 <# 80
      
      ' >> 4 for 16-bit scroll precision, >> 3 for 8 pixel per tile
      if(long[gfxbuffer][spin_gamepad]&NES_LEFT)
        if word[long[gfxbuffer][spin_horizontal_scroll]][193] => 16 
          long[gfxbuffer][spin_tv_communication] := 2 ' make background scroll left (handled from TV driver cog)
          word[long[gfxbuffer][spin_horizontal_scroll]][193] -= 16   ' foreground
          temp1 := (word[long[gfxbuffer][spin_horizontal_scroll]][193] >> 7) ' Get current starting horizontal tile pos
          map_stream_ptr := SRAM_Stream_map_addr + ((temp1 & CONSTANT(MapWidth-1)) << 5) ' Loop stream map around 'MapWidth' tiles wide
          fsrw.readSRAM2d(map_stream_ptr, @foreground[(temp1)&31], 1, 31, 0, 32)
           
      if(long[gfxbuffer][spin_gamepad]&NES_RIGHT)
        if word[long[gfxbuffer][spin_horizontal_scroll]][193] < CONSTANT((MapWidth-31)<<7)
          long[gfxbuffer][spin_tv_communication] := 1 ' make background scroll right (handled from TV driver cog)
          word[long[gfxbuffer][spin_horizontal_scroll]][193] += 16   ' foreground
          temp1 := (word[long[gfxbuffer][spin_horizontal_scroll]][193] >> 7) + 31 ' Get current starting horizontal tile pos
          map_stream_ptr := SRAM_Stream_map_addr + (((temp1) & CONSTANT(MapWidth-1)) << 5) ' Loop stream map around 'MapWidth' tiles wide
          fsrw.readSRAM2d(map_stream_ptr, @foreground[(temp1)&31], 1, 31, 0, 32)
    
    sprite_anim_delay++
    if(sprite_anim_delay == 6)
      sprite_anim_delay := 0
      sprite_anim_ptr += 1024       ' next frame
      if(sprite_anim_ptr => CONSTANT(FLASH_player_sprite_addr + 8192))
        sprite_anim_ptr := FLASH_player_sprite_addr
      'fsrw.readSram(sprite_anim_ptr, @sprite00, 1024)
      fsrw.readFLASH2d(sprite_anim_ptr, sprites[CONSTANT(3 + 0*4)], 32, 0, sprite_dir, CONSTANT(32*32))
    elseif (sprite_anim_delay == 5)
      fsrw.readFLASH2d(sprite_anim_ptr, sprites[CONSTANT(3 + 1*4)], 32, 0, 0, CONSTANT(32*32))
    elseif (sprite_anim_delay == 4)
      fsrw.readFLASH2d(sprite_anim_ptr, sprites[CONSTANT(3 + 2*4)], 32, 0, 1, CONSTANT(32*32))
    elseif (sprite_anim_delay == 3)
      sprite16_anim_ptr += 256
      if sprite16_anim_ptr => CONSTANT(FLASH_sprite16x16_addr + 3072)
        sprite16_anim_ptr := FLASH_sprite16x16_addr
      fsrw.readFLASH2d(sprite16_anim_ptr, sprites[CONSTANT(3 + 3*4)], 16, 0, 0, CONSTANT(16*16))
      fsrw.readFLASH2d(sprite16_anim_ptr+256, sprites[CONSTANT(3 + 4*4)], 16, 0, 0, CONSTANT(16*16))
      fsrw.readFLASH2d(sprite16_anim_ptr+512, sprites[CONSTANT(3 + 5*4)], 16, 0, 0, CONSTANT(16*16))
    elseif (sprite_anim_delay == 2)
      fsrw.readFLASH2d(sprite16_anim_ptr+768, sprites[CONSTANT(3 + 6*4)], 16, 0, 0, CONSTANT(16*16))
      fsrw.readFLASH2d(sprite16_anim_ptr+1024, sprites[CONSTANT(3 + 7*4)], 16, 0, 0, CONSTANT(16*16))
      fsrw.readFLASH2d(sprite16_anim_ptr+1280, sprites[CONSTANT(3 + 8*4)], 16, 0, 0, CONSTANT(16*16))
      'fsrw.readSRAM2d(sprite8_anim_ptr+64+(FrameCounter&64), sprites[CONSTANT(3 + 11*4)], 8, 0, 0, CONSTANT(8*8))
    elseif (sprite_anim_delay == 1)
      ' animated tiles: copy one row of 16 tiles into the tileset, replacing tiles at row 11 (#176..#191) in tiles[]
      fsrw.readSRAM(anim_tiles_ptr, CONSTANT(_memstart + @tiles + 176*64), CONSTANT(64*16))
      anim_tiles_ptr += 1024      ' Move to next tile in animation   
      if(anim_tiles_ptr => CONSTANT(SRAM_Anim_tiles_addr + 16384))
        anim_tiles_ptr := SRAM_Anim_tiles_addr

    if (FrameCounter&1)
      fsrw.readFLASH2d(sprite8_anim_ptr+0+(FrameCounter&2)<<5, sprites[CONSTANT(3 + 9*4)], 8, 0, 0, CONSTANT(8*8))
    else
      fsrw.readFLASH2d(sprite8_anim_ptr+128+(FrameCounter&2)<<5, sprites[CONSTANT(3 + 10*4)], 8, 0, 0, CONSTANT(8*8))
      fsrw.readFLASH2d(sprite8_anim_ptr+256+(FrameCounter&2)<<5, sprites[CONSTANT(3 + 11*4)], 8, 0, 0, CONSTANT(8*8))

    ' animate bullet around the screen
    repeat i from 9 to CONSTANT(gfx#MAX_NB_SPRITE-1)
      temp1 := i<<2
      sprites[temp1+0] += sprite_dirx[i]
      if sprites[temp1+0] > CONSTANT(250<<7) or sprites[temp1+0] < CONSTANT(5<<7)
         sprite_dirx[i] := -sprite_dirx[i]
      sprites[temp1+1] += sprite_diry[i]  
      if sprites[temp1+1] > CONSTANT(187<<7) or sprites[temp1+1] < CONSTANT(21<<7)
         sprite_diry[i] := -sprite_diry[i]
               
    FrameCounter++
  ' Repeat forever
}

PUB PrepareVar
  AutoFireDelay[0] := 16
  AutoFireDelay[1] := 32
  AutoFireDelay[2] := 32
   
  HMoveSpeed[0] := 140
  HMoveSpeed[1] := 110
  HMoveSpeed[2] := 100
   
  JumpPowerStart[0] := 237  ' initial jump power: if just 'tapping' the jump button
  JumpPowerBoost[0] := 14 ' boost power: when holding jump button while jumping
  JumpPowerStart[1] := 255
  JumpPowerBoost[1] := 12
  JumpPowerStart[2] := 255
  JumpPowerBoost[2] := 12

  MapWidth := 128

  SetGameState(0)  ' start directly into state 0 since intro movie was playing during flash booting

' Perform one frame update for the currently selected game state
PUB processGameState
  GameStateTimer++

  case GameState
    0:
      statePrepareScreen
    1:
      stateGameplay
    2:
      stateIntroMovie
    3:
      stateGameOverMovie
      
PUB SetSprite(num, xpos, ypos, size)
  spritesFrame[num] := 0
  if num =< 2
    spritesFlip[num] := 0
  spritesPosX[num] := xpos ' 7-bit fixed point
  spritesPosY[num] := ypos ' 7-bit fixed point
  sprites[2 + (num<<2)] := size         ' width-shift<<8 + height
                                        ' Size also used for visibility: if 0=hide sprite

PUB ClearForeground(value) | i
  repeat i from 0 to 1023
    foreground[i] := value

PUB SetVisibleHUD(hud)
  CurrentHud := hud
  HudSlide := 32 

PUB SetGameState(state) | i
  GameState := state
  GameStateTimer := 0

  case GameState
    0: ' state 0 = prepare screen / display remaining life
      ResetBGScroll
      ' Clear out foreground completely
      ClearForeground(186)
      hscroll193 := 0        ' horizontal scroll start pos (on the left)
      vscroll193 := 0        ' vertical scroll start pos
       
      HitPoint[0] := 6
      HitPointCache := 0
      HitPointBlink := 0
       
      GameTimer := 300 ' 300 seconds
      GameTimerCache := -1
      GameTimerFrame := 0 ' accumulator for 60 frame
       
      Score := 0
      ScoreCache := -1
       
      SetVisibleHUD(0)
       
      ' Clear all sprites
      ClearAllSprites
       
      foreground[CONSTANT(68)] := 56
      foreground[CONSTANT(68+23-1)] := 56
      foreground[CONSTANT(68+64)] := 56
      foreground[CONSTANT(68+23+64-1)] := 56
      foreground[CONSTANT(68+32)] := 59
      foreground[CONSTANT(68+23+32-1)] := 59
      repeat i from 1 To CONSTANT(23-2)
        foreground[i + CONSTANT(68)] := 72
        foreground[i + CONSTANT(32 + 68)] := 1
        foreground[i + CONSTANT(64 + 68)] := 42

      displayString(CONSTANT(@foreground+_memstart + 68 + 32 + 2), string("PREPARE>FOR>COMBAT@"))
      'displayString(CONSTANT(68 + 96 + 2), string("LIFE>LEFT>"))
       
      repeat i from 0 To CONSTANT(127)
        SpawnTiles[i] := 255

      ComputerHacked := 0
       
    1: ' state 1: begin game play
      hscroll193 := 0        ' horizontal scroll start pos (on the left)
      vscroll193 := 80       ' vertical scroll start pos (at bottom)
       
      currentanim[0] := 0
      AnimUpdateImportant[0] := False
      AICounter[0] := 0
       
      JumpRelease[0] := 0
      JumpCounter[0] := 0
      JumpSpeed[0] := 0
       
      InitSprites
      ReloadForeground
      'LoadBgMap("remmap.MAP")
      GameTimerCache := -1 ' force redraw of game timer
      
PUB ClearAllSprites | i
  repeat i from 0 To CONSTANT(GFX#MAX_NB_SPRITE-1)
    SetSprite(i, 0, 0, 0)

PUB InitSprites
  ' make all sprite invisible
  ClearAllSprites
   
  ' This defines a sprite at coordinate 70,100, of size 16x16, using pixel data from @cursor000,
  ' and displaying the 3rd animation frame
  SetSprite(0, 21 << 7, 120 << 7, (5 << 8 + 32))
  ' temp debug: copy sprite 0 because its buffer is shared by other temporary stuff
  'fsrw.readFLASH2d(FLASH_player_sprite_addr, sprites[CONSTANT(3 + 0*4)], 32, 0, 0, CONSTANT(32*32))
  ' removed: sprite should get automatically corrected by AnimateSprite
  
  spritesHotX[0] := 2
  spritesHotX[1] := 2
  spritesHotX[2] := 2
   
  ' Sprites  0 to  2 are 32x32  (total: 3 sprites)
  SetSprite(1, 40<<7, 120<<7, (5 << 8 + 32)) ' sprite 1 is ruffian enemy
  SetSprite(2, 140<<7, 120<<7, (5 << 8 + 32)) ' sprite 2 is another ruffian enemy
  ' Sprites  3 to 8 are 16x16  (total: 6 sprites)
  SetSprite(3, 220<<7, 140<<7, (4 << 8 + 16)) ' sprite 3 will be used for balloon pop-up
   
  ' Sprites 9 to 23 are 8x8    (total: 15 sprites)

PUB BGScrollRight
  long[gfxbuffer][spin_tv_communication] := 1 ' make background scroll right (handled from TV driver cog)

PUB BGScrollLeft
  long[gfxbuffer][spin_tv_communication] := 2 ' make background scroll left (handled from TV driver cog)

PUB statePrepareScreen
  BGScrollRight
  ' debug accelerator: skip part of intro
  'If GameStateTimer < 150 Then GameStateTimer = 150
   
  If GameStateTimer < 16
    temp1 := GameStateTimer << 1
    vscroll194 := 88-temp1 ' foreground start scanline
    hscroll194 := 83+temp1' foreground stop scanline   
    Return
   
  If GameStateTimer > 120 And GameStateTimer < 137
    temp1 := (16-(GameStateTimer-120)) << 1
    vscroll194 := 88-temp1 ' foreground start scanline
    hscroll194 := 83+temp1' foreground stop scanline   
    Return
   
  If GameStateTimer == 150
    ClearForeground(0)
    vscroll194 := 16
    hscroll194 := 191
    vscroll193 := 80       ' vertical scroll start pos (at bottom)
    Return
   
  If GameStateTimer > 150 And GameStateTimer < 183
    temp1 := (GameStateTimer-151)
    hscroll193 := ((temp1 << 3) - 248) << 4
    StreamColumn(temp1, temp1)
    Return
   
  If GameStateTimer == 183
    SetGameState(1)

PUB StreamColumn(src, dst) ' Stream one column on foreground streamed map into the 'foreground screen map'
  ' Loop stream map around 'MapWidth' tiles wide
  fsrw.readSRAM2d(SRAM_Stream_map_addr + ((src & (MapWidth-1)) << 5), @foreground[dst&31], 1, 31, 0, 32)

PUB stateGameplay
  checkInput
  ' After a short pause, game over movie
  If GameStateTimer == 120
    SetGameState(3)
  'hex_display_word(CONSTANT(24+0*gfx#MAPSIZEX), FrameCounter)

PUB checkInput
  repeat WhichSprite from 0 to 2
    WhichSprite4 := WhichSprite<<2
    ProcessInput
    UpdateAnim

  AnimateBullet
  CheckHeroDamage
   
  ' Always follow sprite 0
  WhichSprite := 0
  WhichSprite4 := 0
  ScrollOnSprite
   
  ' Update the game countdown timer
  updateGameTimer

PUB ProcessInput | x,y,desiredanim,FloorLevel
  ' If this sprite is invisible, skip it:
  ' except for hero: he will flicker during damage..
  If WhichSprite <> 0
    If sprites[WhichSprite4] == 0
      Return
   
  ' First, we need to check if this sprite is out of screen:
  ' if so, it cannot move because no collision detection are available outside the visible screen
  x := ((spritesPosX[WhichSprite]+64) ~> 7)
  y := ((spritesPosY[WhichSprite]+64) ~> 7)
  ' If greatly out of screen (about 160 pixel), kill this enemy
  If x < -160 Or x > 400 Or y > 250
    HitPoint[WhichSprite] := 0 ' kill enemy
    sprites[WhichSprite4] := 0 ' turn sprite invisible
    'Print "Enemy id="+WhichSprite+" pos="+x+","+y+" greatly out of screen: auto-killed."
    Return

  ' If slightly out of screen, make sprite immobile
  If x < -30 Or x > 270
    Return
   
  ' if sprite is exploding or dead, skip input processing to prevent movement
  If currentanim[WhichSprite] == 255
    Return
   
  Controls := 0
  CheckKeys

  If Controls & CtrlJump
    ' Initial jump
    ' jump is allowed only if user has release the jump button 
    If JumpRelease[WhichSprite] == 1 And JumpCounter[WhichSprite] == 0
      JumpRelease[WhichSprite] := 0
      JumpCounter[WhichSprite] := 1
      JumpSpeed[WhichSprite] := -JumpPowerStart[WhichSprite] ' initial jump power: if just 'tapping' the jump button
    ElseIf jumpCounter[WhichSprite] > 5 And jumpCounter[WhichSprite] < 15 ' number of frame allowed for 'boosting jump' by holding jump button
      JumpSpeed[WhichSprite] -= JumpPowerBoost[WhichSprite] ' boost power
  Else
    ' only allow jumprelease when falling (or already on floor)
    If JumpSpeed[WhichSprite] => 0
      JumpRelease[WhichSprite] := 1 ' re-allow jumping (user release the jump)
   
  ' first phase: make sprite flip
  If Controls & CtrlLeft
    spritesFlip[WhichSprite] := 1    
  ElseIf Controls & CtrlRight
    spritesFlip[WhichSprite] := 0

  CheckShoot
   
  ' if moving left or right, check for 'wall collision'
  If Controls & (CONSTANT(CtrlLeft | CtrlRight))
    If CheckCollisionSprite <> 0
      ' cancel movement: a collision was detected
      Controls := Controls & (!CONSTANT(CtrlLeft | CtrlRight))
   
  ' second phase: move if allowed
  If Controls & CtrlLeft
    MoveSpriteDelta(-HMoveSpeed[WhichSprite], 0)
  ElseIf Controls & CtrlRight
    MoveSpriteDelta(HMoveSpeed[WhichSprite], 0)

  desiredanim := 1 ' default: run anim

  ' if not moving left or right, transit to idle anim
  If Not (Controls & CONSTANT(CtrlLeft | CtrlRight))
    ' if running on frame 1, then transit back to frame 0 before idling
    temp1 := (spritesFrame[WhichSprite] >> 10) & 15
    If temp1 == 1
      desiredanim := 5
    ElseIf temp1 == 5
      desiredanim := 6
    ElseIf Not (temp1 == 2 Or temp1 == 3 Or temp1 == 6 Or temp1 == 7)
      desiredanim := 0

  ' If jumping
  If JumpCounter[WhichSprite] > 0
    ' if going up, check for ceiling collision
    If JumpSpeed[WhichSprite] < 0
      temp1 := GetFloorSprite(33, 0) ' ceiling collision
      If temp1 <> 0
        jumpSpeed[WhichSprite] := 24
     
    FloorLevel := GetFloorSprite(0, 255 << 7) ' floor collision
     
    jumpCounter[WhichSprite] += 1
    If JumpSpeed[WhichSprite] < 400 ' maximum fall speed
      JumpSpeed[WhichSprite] += 10 ' gravity speed
   
    MoveSpriteDelta(0, JumpSpeed[WhichSprite])
     
    ' when falling, check for floor below player
    If JumpSpeed[WhichSprite] > 0
      desiredanim := 3 ' fall anim
      ' if floor level reached, then 'snap' on it, we have landed
      If spritesPosY[WhichSprite] > FloorLevel
        spritesPosY[WhichSprite] := FloorLevel
        JumpSpeed[WhichSprite] := 0
        JumpCounter[WhichSprite] := 0
        currentanim[WhichSprite] := 4 ' land anim
        AnimUpdateImportant[WhichSprite] := True ' enforce this frame to show
    Else
      desiredanim := 2 ' jump anim
  Else 
    FloorLevel := GetFloorSprite(8, 255 << 7) ' autoclimb
    ' if walking on the ground
    ' if floorlevel is slightly above, make player auto-climb
    If spritesPosY[WhichSprite] > FloorLevel + 150
      spritesPosY[WhichSprite] -= 150
    ' if no floor beneath player, make it fall
    ElseIf spritesPosY[WhichSprite] < FloorLevel
      jumpCounter[WhichSprite] := 99
      JumpSpeed[WhichSprite] := 10
      desiredanim := 3 ' fall anim
  
  If WhichSprite == 0
    CheckSpecialAction
    
  If Not AnimUpdateImportant[WhichSprite]
    currentanim[WhichSprite] := desiredanim

PUB CheckShoot
  If Controls & CtrlShoot
    If ShootCounter[WhichSprite] == 0
      SpawnBullet
    Else
      ShootCounter[WhichSprite] --
  Else
    ShootCounter[WhichSprite] := 0

PUB SpawnBullet | bullet,bulletSizeIndex,yoffset,xoffset,pxoffset,bulletdirIndex
  ' Find a free bullet sprite
  repeat bullet from SPRITE_BULLET_FIRST To SPRITE_BULLET_LAST
    bulletSizeIndex := 2 + bullet << 2
     
    If sprites[bulletSizeIndex] <> 0
      next
     
    yoffset := 12
    xoffset := 10
    pxoffset := 3
    bulletdirIndex := bullet-SPRITE_BULLET_FIRST
     
    SpritesDirX[bulletdirIndex] := 85
     
    If spritesFlip[WhichSprite] == 1
      xoffset := -xoffset
      pxoffset := -pxoffset
      SpritesDirX[bulletdirIndex] := -SpritesDirX[bulletdirIndex]
     
    SpritesPosX[bullet] := SpritesPosX[WhichSprite] + ((CONSTANT((32 >> 1) - 4) + xoffset) << 7)
    SpritesPosY[bullet] := SpritesPosY[WhichSprite] + (yoffset << 7)
     
    If Controls & CtrlUp
      SpritesDirY[bulletdirIndex] := -50
      SpritesPosY[bullet] -= 7 << 7
    ElseIf Controls & CtrlDown
      SpritesDirY[bulletdirIndex] := 50
      SpritesPosY[bullet] += 8 << 7
    Else
      SpritesDirY[bulletdirIndex] := 0
     
    ' Make a semi-random 'spread'
    SpritesDirY[bulletdirIndex] += ((FrameCounter & 7)-4) 
     
    sprites[bulletSizeIndex] := 8 + 3 << 8 ' enable sprite visibility
    ShootCounter[WhichSprite] := AutoFireDelay[WhichSprite]
     
    ' Spawn a 'smoke poff' always using sprite 31
    SpritesPosX[9] := SpritesPosX[bullet] + (pxoffset << 7)
    SpritesPosY[9] := SpritesPosY[bullet]
    spritesFrame[9] := 7 << 6 ' start anim at frame 8 (-1 becuase anim will auto-increment)
    sprites[2 + 9 << 2] := 8 + 3 << 8 ' visibility
    quit
  
PUB CheckCollisionSprite | x,y,x1,x2,hoffset,htile,htile2,i,i2,i3
  ' Default 'hot spot' is center horizontal, bottom vertical
  x := ((spritesPosX[WhichSprite]+64) ~> 7) + CONSTANT(32 >> 1)
  y := ((spritesPosY[WhichSprite]+64) ~> 7) + CONSTANT(32-1)
   
  ' unknown magic value: 12
  hoffset := 12
  If spritesFlip[WhichSprite] == 1
    hoffset := -hoffset
   
  x1 := x + hoffset
  x2 := x + (hoffset ~> 2)
     
  htile := ((hscroll193 >> 4 + x1) >> 3) & 31
  htile2 := ((hscroll193 >> 4 + x2) >> 3) & 31
   
  If x1 < 0 Or x1 > 248
    htile := 255 ' if sprite X is off-screen, invalidate collision
  If x2 < 0 Or x2 > 248
    htile2 := 255 ' if sprite X is off-screen, invalidate collision
   
  i := ((vscroll193 + y - 24) >> 3)
  i2 := (i+1)
  i3 := (i+2)
   
  If IsCollisionAt(htile, i)
    Return 1
  If IsCollisionAt(htile, i2)
    Return 1
  If IsCollisionAt(htile2, i3)
    Return 1
  Return 0

PUB GetFloorSprite(verticaloffset, maximumPos) | x,y,htile,htile2,i
  ' Default 'hot spot' is center horizontal, bottom vertical
  x := ((spritesPosX[WhichSprite]+64) ~> 7) + CONSTANT(32 >> 1)
  y := ((spritesPosY[WhichSprite]+64) ~> 7) + CONSTANT(32-1)
   
  ' unknown magic value: -4 ??
  htile := ((hscroll193 >> 4 + x - 4) >> 3) & 31
  htile2 := (htile+1) & 31
   
  ' calculate which tile the sprite is colliding on
  i := ((vscroll193 + y - verticaloffset) >> 3)
   
  If IsCollisionAt(htile, i) Or IsCollisionAt(htile2, i)
    ' +4 to make sprite 'sink' a bit into the floor
    Return ((i << 3) - vscroll193 + CONSTANT(-32 + 4)) << 7
   
  ' special magic: if colliding with a floor, perform auto-climb for 1 tile 'staircase'
  If verticaloffset == 8 And i < 31
    i += 1
    If IsCollisionAt(htile, i) Or IsCollisionAt(htile2, i)
      ' +4 to make sprite 'sink' a bit into the floor
      Return ((i << 3) - vscroll193 + CONSTANT(-32 + 4)) << 7
   
  Return maximumPos  ' if no floor/ceiling, return the topmost/bottomest position

PUB IsCollisionAt(x, y)
  If x > 31
    Return 1 ' if 'x' is off-screen, block with always collision
  If y > 31
    Return 0 ' no collision if outside screen
  Return IsTileCollision(foreground[x + y << 5])

PUB IsTileCollision(tile)
  If tile => 144 And tile =< 155
    Return 1
  Return 0

PUB CheckSpecialAction       ' Special hero function
  GameStateTimer := 0 ' always reset game timer during gameplay 


PUB CheckKeys
  If WhichSprite == 0
    CheckPlayerKeys
  ElseIf WhichSprite == 1 Or WhichSprite == 2
    CheckRuffianKeys

' Normal player control
PUB CheckPlayerKeys
  Controls := SpinGamepad ' simplified by mapping Controls exactly like SpinGamepad
  {if(SpinGamepad&NES_LEFT)
    Controls |= CtrlLeft
    
  if(SpinGamepad&NES_RIGHT)
    Controls |= CtrlRight       
   
  if(SpinGamepad&NES_A)
    Controls |= CtrlJump
    
  if(SpinGamepad&NES_B)
    Controls |= CtrlShoot
   
  if(SpinGamepad&NES_UP)
    Controls |= CtrlUp
    
  if(SpinGamepad&NES_DOWN)
    Controls |= CtrlDown}       

' Ruffian enemy: move toward player (which is sprite 0)
PUB CheckRuffianKeys
  ' debug: make ruffian#2 move back n forth endlessly
  if WhichSprite == 2
    if FrameCounter & 128 < 64
      Controls := CtrlRight
    else
      Controls := CtrlLeft
  elseif WhichSprite == 1
    if (FrameCounter & 4) == SpriteAnimDelay
      Controls := CtrlJump
  ' nothing

PUB UpdateAnim | startFrame, endFrame, frame
  If sprites[WhichSprite4] == 0
    Return
   
  ' anim number 0..127 are looping, normal control anim with up/down look
  ' anim number 128..255 are play-once.
  ' Anim 255 is used For death/explode animation and sprite will be killed after playback
  case currentanim[WhichSprite]
    0: ' idle anim
      startFrame := 8
      endFrame := 11
    1: ' walk anim
      startFrame := 0
      endFrame := 7
    2: ' jump anim
      startFrame := 12
      endFrame := 12
    3: ' fall anim
      startFrame := 13
      endFrame := 13
    4: ' land anim
      startFrame := 14
      endFrame := 14
    5: ' run / idle transit frame 0
      startFrame := 0
      endFrame := 0
    6: ' run / idle transit frame 4
      startFrame := 4
      endFrame := 4
    255: ' explode
      startFrame := 48
      endFrame := 53
   
  If SpriteAnimDelay == 6
    frame := spritesFrame[WhichSprite] >> 10
    If currentanim[WhichSprite] < 128
      frame &= 15
    Else
      ' if animation has reached lastframe, then kill sprite
      If frame == endFrame
        sprites[WhichSprite4] := 0 ' turn sprite invisible
        Return
     
    frame++
    If frame < startFrame Or frame > endFrame
      frame := startFrame
            
    If currentanim[WhichSprite] < 128
      If WhichSprite == 0 
        If Controls & CtrlUp
          frame += 16
        ElseIf Controls & CtrlDown 
          frame += 32
     
    AnimUpdateImportant[WhichSprite] := False ' we have reach (displayed) one frame of the requested anim            
    spritesFrame[WhichSprite] := frame << 10
   
PUB AnimateBullet
  ' nothing

PUB CheckHeroDamage
  ' nothing

PUB ScrollOnSprite | dir,x,y,scrollx,scrolly
  dir := spritesFlip[WhichSprite]
  ' Default 'hot spot' is center horizontal, bottom vertical
  x := ((spritesPosX[WhichSprite]+64) ~> 7) + (16)
  y := ((spritesPosY[WhichSprite]+64) ~> 7) + (16)
   
  ' dir=0 when facing right ->
  ' dir=1 when facing left <-
  ' hzone = size of horizontal 'no scroll window'
   
  scrollx := 0
  scrolly := 0
   
  If dir == 0 And x > CONSTANT(128-30)
    scrollx := HScrollRight
  ElseIf dir == 1 And x < CONSTANT(128+30)
    scrollx := HScrollLeft
   
  If y > CONSTANT(96+6)
    scrolly := VScrollDowm(y - CONSTANT(96+6))
  ElseIf y < CONSTANT(96-6)
    scrolly := VScrollUp(CONSTANT(96-6) - y)
   
  MoveSpritesDelta(scrollx, scrolly)

PUB MoveSpritesDelta(dx, dy)
  repeat WhichSprite from 0 to CONSTANT(GFX#MAX_NB_SPRITE-1)
    MoveSpriteDelta(dx, dy)

PUB MoveSpriteDelta(dx, dy)
  spritesPosX[WhichSprite] += dx  ' 7-bit fixed point
  spritesPosY[WhichSprite] += dy  ' 7-bit fixed point

PUB VScrollUp(distance)
  If distance > vscroll193
    distance := vscroll193
  ' block vertical scroll at '0'
  vscroll193 -= distance 
  Return distance << 7

PUB VScrollDowm(distance)
  If distance > 80-vscroll193
    distance := 80-vscroll193
  ' block vertical scroll at 32*8 = 256 - (192 - 16) = 80
  vscroll193 += distance 
  Return (-distance) << 7

PUB HScrollRight | htile
  If hscroll193 < ((MapWidth-31) << 7)
    hscroll193 += 16
     
    BGScrollRight
     
    htile := (hscroll193 >> 7) + 31 ' Get current starting horizontal tile pos
    StreamColumn(htile, htile) ' values will be 'anded' by StreamColumn     
    Return -128

  Return 0

PUB HScrollLeft | htile
  If hscroll193 => 16
    hscroll193 -= 16
     
    BGScrollLeft
    htile := (hscroll193 >> 7) ' Get current starting horizontal tile pos
    StreamColumn(htile, htile) ' values will be 'anded' by StreamColumn     
    Return 128
    
  Return 0

PUB updateGameTimer
  GameTimerFrame++
  If GameTimerFrame == 60
    If GameTimer > 0
      GameTimer--
    GameTimerFrame := 0

PUB stateIntroMovie
  CutsceneMode(movie_cut)
  SetGameState(0)

PUB stateGameOverMovie
  CutsceneMode(gameover_cut)
  SetGameState(2)

PUB updateHUD
  ' nothing

PUB animatetiles ' Perform tile animation
  If SpriteAnimDelay <> 2
    return
    
  anim_tiles_ptr += 1024      ' Move to next tile in animation   
  if(anim_tiles_ptr => CONSTANT(SRAM_Anim_tiles_addr + 16384))
    anim_tiles_ptr := SRAM_Anim_tiles_addr

  ' animated tiles: copy one row of 16 tiles into the tileset, replacing tiles at row 11 (#176..#191) in tiles[]
  fsrw.readSRAM(anim_tiles_ptr, CONSTANT(_memstart + @tiles + 176*64), CONSTANT(64*16))

PUB rendersprites | i
  MoveSprites
  repeat i from 0 to CONSTANT(GFX#MAX_NB_SPRITE-1)
    rendersprite(i)
  AnimateSprites

PUB rendersprite(i) | ishift     
  ishift := i << 2

  ' Update sprite screen position:
  sprites[(0 + ishift)] := spritesPosX[i]
  ' Apply vertical HUD offset
  sprites[(1 + ishift)] := spritesPosY[i] + ((vscroll194 - FLOOROFFSET) << 7)
   
  ' Apply horizontal hotspot offset: this only applies to sprite 0,1,2
  if i =< 2 
    If spritesFlip[i] == 1
      sprites[(0 + ishift)] -= (spritesHotX[i] << 7)
    Else
      sprites[(0 + ishift)] += (spritesHotX[i] << 7)
     
PUB MoveSprites | i
  repeat i from SPRITE_BULLET_FIRST to SPRITE_BULLET_LAST
    SpritesPosX[i] += (~SpritesDirX[i-SPRITE_BULLET_FIRST]) << 2
    SpritesPosY[i] += (~SpritesDirY[i-SPRITE_BULLET_FIRST]) << 2

PUB AnimateSprites | i ' make sprite animate by streaming from FLASH to internal ram the requested frame
  Case SpriteAnimDelay
    6: ' animate sprite 0 (hero)
      fsrw.readFLASH2d(FLASH_player_sprite_addr +  spritesFrame[0], sprites[CONSTANT(3 + 0*4)], 32, 0, spritesFlip[0], CONSTANT(32*32))
    5: ' animate sprite 1 (ruffian) 
      fsrw.readFLASH2d(FLASH_ruffian_sprite_addr +  spritesFrame[1], sprites[CONSTANT(3 + 1*4)], 32, 0, spritesFlip[1], CONSTANT(32*32))
    4: ' animate sprite 2 (ruffian)
      fsrw.readFLASH2d(FLASH_ruffian_sprite_addr +  spritesFrame[2], sprites[CONSTANT(3 + 2*4)], 32, 0, spritesFlip[2], CONSTANT(32*32))
    3: ' animate sprite 3 (action bubble)
      fsrw.readFLASH2d(FLASH_sprite16x16_addr +  spritesFrame[3], sprites[CONSTANT(3 + 3*4)], 16, 0, 0, CONSTANT(16*16))
    ' 2 is used for animated tiles

  ' update all bullets at 1/2 framerate   
  if FrameCounter&1
    fsrw.readFLASH2d(FLASH_sprite8x8_addr+0+(FrameCounter&2)<<5, sprites[CONSTANT(3 + 11*4)], 8, 0, 0, CONSTANT(8*8))
   
  AnimateParticle(10, 12, 16)
  AnimateParticle(9, 8, 12)

PUB AnimateParticle(sprite, animStart, animEnd) | frame
  temp1 := 2 + sprite << 2
  If sprites[temp1] == 0
    Return
   
  If FrameCounter & 3
    Return
   
  ' animate particle: anim from frame 'animStart' to 'animEnd'
  frame := spritesFrame[sprite] >> 6
  frame++
  If frame == animEnd ' hide sprite when anim is finished
    sprites[temp1] := 0
    frame := animStart ' prepare sprite for next particle

  spritesFrame[sprite] := frame << 6
  'fsrw.readFLASH2d(sprite8_anim_ptr+0+(FrameCounter&2)<<5, sprites[CONSTANT(3 + 9*4)], 8, 0, 0, CONSTANT(8*8))
   
PUB SwitchGfxEngine(mode)
  if mode == 0
    temp2 := SRAM_GFX_tile_mode_addr
  else
    temp2 := SRAM_GFX_cutscene_mode_addr  

 ' Turn off screen before:
  word[long[gfxbuffer][spin_horizontal_scroll]][194] := 511
  wait_tv_vbl
  
   ' overwrite 'tilemap and foreground' with cog code stored in SRAM.. 
  fsrw.readSram(temp2, @tilemap, 2048)
  
  temp2 := cogid_gfx[0] ' indicator for 'first time': we need to 'cognew'. The successive calls will use 'coginit' instead
  temp1 := 0
  repeat
     long[gfxbuffer][spin_cog] := temp1
    ' Warning: we can't ASSUME that gfx cogs always start at cogid 3 and up, since when launched from JT boot loader,
    ' the SPIN running cog might be a different number
    if temp2 == 255
      cogid_gfx[temp1] := cognew(@tilemap, gfxbuffer)
    else
      coginit(cogid_gfx[temp1], @tilemap, gfxbuffer)
    repeat
    while long[gfxbuffer][spin_cog] <> 255
    temp1++
  until temp1 == gfx#GFX_COG_TOTAL

  ' Clear foreground and background to prevent showing garbage
  repeat temp1 from 0 to 511 ' 2048 bytes / 4 = 512 longs
    tilemap[temp1] := 0
    
  long[gfxbuffer][spin_cog] := 511                  ' Magic value: start all GFX cogs
  
PUB CutsceneMode(whichMovie)
  ' During Cutscenemode, the GFX engine is set into a different state which uncompressed a fullscreen 256x192
  ' image. The HUB RAM buffer used for the compressed picture is overwriting the 'tiles' buffer.
  ' Thus, the GFX engine must be turned off temporarily while preparing the data
  ' Then the GFX engine must be switched into the 'cutscene mode' render algorithm.
  ' Same thing when going back to normal tile mode.

  ' Black the screen completely
  SwitchGfxEngine(1)
  
  SpinGamepad := NES_START
  'DisplayString(CONSTANT(8+0*gfx#MAPSIZEX), string("CUTSCENE>MODE>"))

  ' Read in movie CLUT (102 bytes) over @tilemap:
  fsrw.seek(moviecol_pal)
  fsrw.pread(@tilemap, moviecol_pal_size)

  fsrw.seek(whichMovie)
  fsrw.pread(@movieLen, 2)
  movie_frame := 0
  ' Compute size of movie frame pointer table: movieLen*2
  ' Read to SRAM the complete movie pointer table (it contains 16-bit size for each individual frames)
  ' Currently we have a maximum of 32KB to store the table, so we have a maximum movie length of 16384 frames (minus the CLUT)
  CopySDtoSRAM_NoSeek(movieLen << 1, SRAM_Movie_Table_addr)
  ' Compute the seek position of the first frame, in order to 'loop' the movie if needed
  movie_seek_loop := whichMovie+2+ (movieLen << 1)

  movie_frame_cnt := cnt-1 ' force first frame to appear instantly

  CutsceneFetchFrame

  word[long[gfxbuffer][spin_horizontal_scroll]][194] := 191 ' enable display

  ' we can't display any text in cutscene mode: no more tiles in memory, and no more tile engine running.. ouch.
  'hex_display_word(CONSTANT(24+1*gfx#MAPSIZEX), movieLen)
  'hex_display_long(CONSTANT(14+1*gfx#MAPSIZEX), temp1)
   
  repeat
    'wait_vbl
    'sound_music_update

    'tv_color := $8D ' yellowish
    'repeat temp1 from 0 to 500
    temp1 := NES_Read_Gamepad
    if temp1&NES_START and not (SpinGamepad&NES_START)
      movie_frame := -1
    SpinGamepad := temp1

    if movie_frame == -1 ' end of movie or user aborted
      SwitchGfxEngine(0)
      RefreshFromSD
      hscroll194 := 191
      return

    CutsceneFetchFrame
          
    if movie_frame == movieLen
      if whichMovie == movie_cut ' the intro movie loops forever until user abort
        movie_frame := 0
        fsrw.seek(movie_seek_loop)
      else
        movie_frame := -1         ' other movie will only play once

    
PUB CutsceneFetchFrame | frame_length
  ' Fetch first frame address from table in SRAM:
  ' Note: 'frame_length' can be 0 for a duplicate frame
  fsrw.readSram(SRAM_Movie_Table_addr + (movie_frame<<1), @frame_length, 2)
  ' Read a RLE packed frame into SRAM: 'temp2' contains the length of the current frame
  'wait_vbl
  ' This copy can take a couple of vbl, we'll need a way to time it in order to play movie at a predictable rate
  'tv_color := $8D ' yellowish
  sound_music_update
  CopySDtoSRAM_NoSeek(frame_length, SRAM_Anim_tiles_addr)

  ' Here we must wait some calculated time before displaying next frame in order to have a fixed and smooth movie rate.
  ' 'cnt' increments every clock cycle, we run at 80Mhz
  ' If movie plays at 10 frame per second:
  ' 80000000 / 10 = 8000000 cnt must elasped before next frame
  ' one NTSC frame = 80000000 / 60 = 1333333 clocks
  ' We should update the music every 1 million clock

  repeat until cnt - movie_frame_cnt => 8000000
    sound_music_update
    
  movie_frame_cnt := cnt   
  
  ' now the crucial part: wait a vbl, then quickly copy from SRAM to HUB:
  ' the trick is that we will be 'racing the beam' since the cutscene driver are rendering the frame in realtime
  ' while we stream a new frame of data, overwriting the same buffer. 
  'wait_vbl
  ' We might need to insert some kind of delay here, in order to let the cutscene driver take a little bit of advance
  ' The problem is that a 'simple frame' could take 1000 bytes, and a very complex frame could take 12000 bytes
  ' So we can't assume what delay should be inserted.. It needs to be calculated.
  'tv_color := $AD
  ' Wait until TV driver reach almost the bottom of the screen
  repeat until long[gfxbuffer][spin_request_scanline] == 180
  'repeat temp1 from 0 to 1400
  'tv_color := $02
  sound_music_update
  
  ' Transfer a complete frame of compressed data over the 'tiles' 12KB buffer
  ' Since the maximum read speed is about 7177 bytes per frame, it can take a little bit more than one complete frame to read the data 
  fsrw.readSram(SRAM_Anim_tiles_addr, @tiles, frame_length)
  sound_music_update
  
  movie_frame++
 
PUB InitSpritePointers | i 
  ' Startup sprite definition.
  ' You don't have to specify sprites here, since they could be defined at runtime in the
  ' assembly code.
  ' Here's how sprite are organised:
  ' Sprite 0 is drawn first, so it will be shown as the farthest
  ' Sprite 31 is drawn last, so it always show on top of everything

  ' X and Y position are stored in 7-bit fixed-point format, using variables spin_sprite_x and spin_sprite_y.
  ' i.e.: If you want to display a sprite at pixel 50,100 on screen,
  '       you'll need to set x and y to CONSTANT(50<<7), CONSTANT(100<<7)
  ' Sprite size is stored in spin_sprite_s. Horizontal width is shifted by 8 bit and is stored as a shift value,
  ' then vertical height is added to it unshifted.
  ' i.e.: To set a 32x50 sprite size, you'll have to known that 32 equal 1 << 5,
  '       so final result would be CONSTANT(5<<8 + 50)
  ' Memory address
  
  ' first sprite (0):
  ' This defines a sprite at coordinate 70,100, of size 32x32, using pixel data from @sprite00
  sprites[CONSTANT(3 + 0*4)] := gfx.RV_getcog_overlap 'CONSTANT(@sprite00+_memstart) '32x32
  sprites[CONSTANT(3 + 1*4)] := snd.RV_getcog_overlap
  sprites[CONSTANT(3 + 2*4)] := tv.RV_getcog_overlap         
  sprites[CONSTANT(3 + 3*4)] := snd.RV_getcog_overlap + 1024 '16x16 
  sprites[CONSTANT(3 + 4*4)] := tv.RV_getcog_overlap + 1024
  sprites[CONSTANT(3 + 5*4)] := sprites[CONSTANT(3 + 6*4)] := sprites[CONSTANT(3 + 7*4)] := sprites[CONSTANT(3 + 8*4)] := sprites[CONSTANT(3 + 4*4)] 
  sprites[CONSTANT(3 + 9*4)] := tv.RV_getcog_overlap + 1280  '8x8
  sprites[CONSTANT(3 + 10*4)] := tv.RV_getcog_overlap + 1344 
  sprites[CONSTANT(3 + 11*4)] := snd.RV_getcog_overlap + 1670 

  repeat i from 12 to CONSTANT(gfx#MAX_NB_SPRITE-1)
    sprites[i<<2+3] := sprites[CONSTANT(3 + 11*4)]
      
PUB wait_tv_vbl
    ' Wait for a VBL using the legacy TV variable 'tv_status'
    'tv_color := $6A             ' green = portion spent in vertical blank + overscan lines
    repeat while tv_status == 1 ' while in vertical blank

    'tv_color := $02             ' black = free portion
    repeat while tv_status == 2 ' while active screen is being drawn
    
    'tv_color := $4B             ' red = used portion by SPIN code

PUB wait_vbl
  ' perform sound and music update
    sound_music_update 
    wait_tv_vbl
    'repeat temp1 from 0 to 50
        
PUB sound_music_update 
  ' NOTE: looping music is not currently implemented since our 32 bytes buffer is streamed from SRAM
  ' We'll need to reset 'music_ptr' when song is completed
  
    'tv_color := $DB             ' show sound driver usage (violet)
    snd.SFX_do_playback
    temp3 := snd.do_playback
    if(temp3 <> 0)
      fsrw.readSram(music_ptr, @music_buffer[32 - temp3], temp3)
      music_ptr += temp3
    ' update music twice for better playback
    temp3 := snd.do_playback
    if(temp3 <> 0)
      fsrw.readSram(music_ptr, @music_buffer[32 - temp3], temp3)
      music_ptr += temp3

PUB MountSD
  snd.stop_song
  DisplayString(CONSTANT(@tilemap+_memstart + 8+0*gfx#MAPSIZEX), string("MOUNTING>SD>>>"))
                                                                        
  repeat
    repeat
      temp1 := \fsrw.mount
      hex_display_word(CONSTANT(22+0*gfx#MAPSIZEX), temp1)
    until temp1 == 0            ' Attempt to mount SD card until it succeed 
     
    if \fsrw.popen(string("PROJECTB.WAD"),"r") <> 0
      DisplayString(CONSTANT(@tilemap+_memstart + 8+0*gfx#MAPSIZEX), string("FILE>NOT>FOUND"))
      next
    else
      quit

PUB RefreshFromSD
  MountSD
     
  DisplayString(CONSTANT(@tilemap+_memstart + 8+0*gfx#MAPSIZEX), string("READING>SD>>>>>"))
  
  ' Fetch tileset
  fsrw.seek(tiles_dat)
  fsrw.pread(@tiles, tiles_dat_size)

  ' Fetch bgscrollspeed, using temporary storage
  fsrw.seek(BGscrollSpeed_dat)
  fsrw.pread(@foreground, BGscrollSpeed_dat_size)
  long[gfxbuffer][spin_tv_communication] := 3
  ResetBGScroll ' do this while tv driver is copying the scroll data
  ' Await 'acknowledge' by TV driver:
  repeat while long[gfxbuffer][spin_tv_communication] == 3

  ' Fetch background map
  fsrw.seek(remmap_map)
  fsrw.pread(@tilemap, remmap_map_size)

  DisplayString(CONSTANT(@tilemap+_memstart + 8+0*gfx#MAPSIZEX), string("COPY>MUSIC>SRAM"))
  CopySDtoSRAM(HDMF_song_bin, HDMF_song_bin_size, 0)
  '(warning: music must NOT exceed 8192 or else other items 'stomp' it)

  ' Copy animated tiles 
  CopySDtoSRAM(anim_tiles_dat, anim_tiles_dat_size, SRAM_Anim_tiles_addr)
  anim_tiles_ptr := SRAM_Anim_tiles_addr

  ' Copy stream map to memory:
  CopySDtoSRAM(stream_map, stream_map_size, SRAM_Stream_map_addr)
  ' Reset scroll position, reset map stream pointer
  hscroll193 := 0
  'map_stream_ptr := SRAM_Stream_map_addr
  'ReloadForeground ' don't do this yet

  DisplayString(CONSTANT(@tilemap+_memstart + 8+0*gfx#MAPSIZEX), string("COPY>COMPLETED>"))

  ' Fetch start of music
  fsrw.readSram(0, @music_buffer[0], 32)
  music_ptr := CONSTANT(SRAM_music_addr + 32)   
  snd.play_song(@music_buffer[0],0)'snd#HDMF__FLAG_LOOP_SONG)
  snd.do_playback

' Load data from Flash, which are loaded once (i.e.: never need reloading)
PUB LoadFlashData
  DisplayString(CONSTANT(@tilemap+_memstart + 8+0*gfx#MAPSIZEX), string("COPY>SPRITES>>>"))
  CopySDtoFLASH(bot_sprite_dat, bot_sprite_dat_size, FLASH_player_sprite_addr)

  CopySDtoFLASH(ruffian_sprite_dat, ruffian_sprite_dat_size, FLASH_ruffian_sprite_addr)
  'sprite_anim_ptr := FLASH_player_sprite_addr

  CopySDtoFLASH(sprite16x16_dat, sprite16x16_dat_size, FLASH_sprite16x16_addr)
  'sprite16_anim_ptr := FLASH_sprite16x16_addr

  CopySDtoFLASH(sprite8x8_dat, sprite8x8_dat_size, FLASH_sprite8x8_addr)
  'sprite8_anim_ptr := FLASH_sprite8x8_addr

PUB ResetBGScroll
  repeat temp1 from 0 to 191
    word[long[gfxbuffer][spin_vertical_scroll]][temp1] := temp1
    word[long[gfxbuffer][spin_horizontal_scroll]][temp1] := 32                  ' Start BG horizontal scroll by half the 6-bit fixed-width precision

PUB ReloadForeground
  ' Then pre-stream columns 0 to 31 before starting
  repeat temp1 from 0 to 31
    StreamColumn(temp1, temp1)
    'fsrw.readSRAM2d(SRAM_Stream_map_addr + (temp1<<5), @foreground[temp1], 1, 31, 0, 32)

PUB CopySDtoSRAM(src_sd_start_pos, file_length, dst_sram_address)
  ' Copy SD data into SRAM
  fsrw.seek(src_sd_start_pos)
  CopySDtoSRAM_NoSeek(file_length, dst_sram_address)

PUB CopySDtoSRAM_NoSeek(file_length, dst_sram_address)
  CopySDtoMemory_Internal(file_length, dst_sram_address, 0)

' perform copy using 'foreground' tilemap as a stratch buffer: 256 bytes will be trashed
PUB CopySDtoMemory_Internal(file_length, dst_address, is_flash)
  ' Copy SD data into SRAM or FLASH
  repeat while file_length > 0
    temp2 := file_length <# 256 
    fsrw.pread(@foreground, temp2) ' read up to 256 byte at a time
    if is_flash
      fsrw.writeData(dst_address, @foreground, temp2)
    else
      fsrw.writeSram(dst_address, @foreground, temp2)
    dst_address += temp2
    file_length -= temp2

PUB CopySDtoFLASH(src_sd_start_pos, file_length, dst_flash_address)
  ' Copy SD data into FLASH (write-once)
  fsrw.seek(src_sd_start_pos)
  CopySDtoMemory_Internal(file_length, dst_flash_address, 1)

'end of main
'---------------------------------------------
PUB hex_display_nibble(pos, value)
  if(value < 10)
    byte[@tilemap+pos] := 32 + value ' Number '0' starts at tile 32
  else
    byte[@tilemap+pos] := CONSTANT(4-10) + value ' 'A' starts at tile 4, minus 10 for adjustment  


PUB hex_display_byte(pos, value)
  hex_display_nibble(pos, (value>>4)&$F)
  hex_display_nibble(pos+1, (value)&$F)

PUB hex_display_word(pos, value)
  hex_display_byte(pos, (value>>8)&$FF)
  hex_display_byte(pos+2, (value)&$FF)

PUB hex_display_long(pos, value)
  hex_display_word(pos, (value>>16)&$FFFF)
  hex_display_word(pos+4, (value)&$FFFF)

PUB DisplayString(pos, text)
  'pos+=@tilemap
  repeat until byte[text] == 0
    byte[pos] := byte[text]+CONSTANT(-65+4)  ' Character 'A' starts at tile #4
    text++
    pos++  
  
' General read gamepad spin function
PUB NES_Read_Gamepad : nes_bits | i
  DIRA [5] := 1 ' output JOY_CLK
  DIRA [4] := 1 ' output JOY_SHLDn
  'DIRA [3] := 0 ' input  JOY_DATAOUT0
  'DIRA [2] := 0 ' input  JOY_DATAOUT1
  DIRA [1] := 1 ' output JOY_5V
  DIRA [0] := 1 ' output JOY_GND
   
  OUTA [1] := 1 ' enable 5V to high
  'OUTA [0] := 0 ' set GND to low
   
  'OUTA [5] := 0 ' JOY_CLK = 0
  'OUTA [4] := 0 ' JOY_SH/LDn = 0
  OUTA [4] := 1 ' JOY_SH/LDn = 1
  OUTA [4] := 0 ' JOY_SH/LDn = 0
  'nes_bits := 0
  'nes_bits := INA[3]' | (INA[2] << 8)
   
  repeat i from 0 to 7
    nes_bits := (nes_bits << 1)
    nes_bits := nes_bits | INA[3] | (INA[2] << 8)
    OUTA [5] := 1 ' JOY_CLK = 1
    OUTA [5] := 0 ' JOY_CLK = 0
   
  nes_bits := (!nes_bits & $FFFF)
  if(nes_bits&$00ff == $00ff) ' controller 0 not plugged in, pretend all buttons are unpressed.
    nes_bits&=$ff00
  if(nes_bits&$ff00 == $ff00) ' controller 1 not plugged in, pretend all buttons are unpressed.
    nes_bits&=$00ff



' Start of assembler section
DAT
' Legacy TV driver stuff. DON'T MODIFY ANYTHING HERE.
tvparams
tv_status                       long    0               '0/1/2 = off/visible/invisible           read-only 
tv_enable                       long    1               '0/? = off/on                            write-only
tv_pins                         long    %001_0101       '%ppmmm = pins for C3                    write-only
tv_mode                         long    %0000           '%ccinp = chroma,interlace,ntsc/pal,swap write-only
tv_color                        long    $02             'background (overscan color)             write-only
tv_ht                           long    16              'horizontal cells                        write-only
tv_vt                           long    12              'vertical cells                          write-only
tv_hx                           long    10              'horizontal cell expansion               write-only
tv_vx                           long    1               'vertical cell expansion                 write-only
tv_ho                           long    0               'horizontal offset                       write-only
tv_vo                           long    0               'vertical offset                         write-only
tv_broadcast                    long    60_000_000'_xinfreq<<4  'broadcast frequency (Hz)                write-only
tv_auralcog                     long    0               'aural fm cog                            write-only
tv_gfxbuffer                    long    0               'RV: gfx buffer address


' Start back ground tilemap definition here (BG map is now streamed)
' Note: tilemap and foreground (both 1024 bytes) must remain contiguous in memory since
' we are using this combined 2KB buffer for gfx/cutscene mode code swap
tilemap      'file "remmap.map"
              byte 0[32*32] ' BG Map size is 32x32
' The foreground is now streamed directly so here is just a reserved space:
' Note: this foreground map will horizontal 'wrap around', making it appear to scroll indefinitely
' But remember than its width is always 32 in memory, so only 'one screen' of tiles are available for gameplay:
' i.e.: ennemis or collision cannot occur with tiles that are scrolled off screen
foreground   'file "foreground.map"
              byte 0[32*32] ' FG Map size is 32x32

' Tiles are present in the 32KB .BIN because we need to display string message to the user in case an SD card is not present
' This area will be overwritten at runtime by reading from SD
' This is a bitmap area of 128 x 96 = 12288 bytes = 192 tiles
' NOTE: in cutscene mode, this memory is shared as a 'frame buffer' for the current compressed RLE movie frame
tiles          file "tiles.dat"