''*************************************
''*  Rem Cutscene Graphic engine v020 *
''*************************************

CON

  GFX_COG_TOTAL = 4
  
  MAX_NB_SPRITE = 32   


  ' Reminder: the screen is 256x192, which gives 32x24 tiles.
  ' Limitation:
  ' MAPSIZEX must be a power of 2.
  ' MAPSIZEY have no restriction, can be from 1 to infinity.
  MAPSIZEX = 32
  MAPSIZEY = 32
  MAPSIZEX_SHIFT = 5    ' MAPSIZEX_SHIFT must be set according with MAPSIZEX

  'Map dimension for FOREGROUND map
  FGSIZEX = 32
  FGSIZEY = 32
  FGSIZEX_SHIFT = 5    ' FGSIZEX_SHIFT must be set according with FGSIZEX
  
' Cutscene mode: here goes the runtime RLE decompressor when GFX engine is switched into cutscene/movie mode:
' Each scanline must uncompress one line of graphic data: i.e.: exactly 256 bytes
' The data starts in the 'tiles' buffer
' For example: if the currentscanline is 0, the data to unpack starts at 'tiles'.
' Then the other cogs must also process this line of data by 'uncompressing it' silently in order to reach the next line of data
' After a line has been emitted to the TV, the cog must then read and skip 'cognumber' numbers of line of data in order to reach
' their relevant next line, since we are emitting with multi-interleaved cog decompressing.
' (Note: this might not be required, in fact one cog could probably uncompress the whole screen by itself)

{PUB start(cognum, paramadr)

'' Start REM engine - starts a cog
  coginit(cognum, @entry, paramadr)
}
pub RV_getcog_overlap
  return @entry
  
PUB RV_getcog_end_of_ram
  return (end_of_cog_ram-1)<<2


DAT

                        org

' Entry
'
entry                   'mov dira, #1 ' enable debug led port
                        ' fetch some parameter, starting at 'par':
                        rdlong cognumber, par ' store cog number (0,1,...)
                        mov t2, #255            ' signal that we acknowledged our cognumber
                        wrlong t2, par           

                        ' Wait here until 'par' points to 511 (magic value to make gfx cogs starts)
wait_startup            rdlong t2, par
                        cmp t2, #511 wz
              if_nz     jmp #wait_startup

                        'movs add_cog_total,   ' store total number of cogs (1,2,...)

                        mov t2, par             ' prefect some pointer to speed up runtime execution:
                        add t2, #4              '   spin_request_scanline = 1
                        mov request_scanline, t2
                        add t2, #4              ' spin_vertical_scroll = 2      
                        rdlong vertical_scroll, t2
                        add t2, #4              ' spin_horizontal_scroll = 3          
                        rdlong horizontal_scroll, t2
                        add t2, #4              ' spin_sprites_adr = 4     
                        rdlong sprite_adr, t2
                        add t2, #4              ' spin_tilemap_adr = 5        
                        rdlong tilemap, t2      ' In cutscene mode, 'tilemap' is used for 'CLUT' table
                        add t2, #4              ' spin_tiles_adr = 6    
                        rdlong tiles, t2        ' In cutscene mode, 'tiles' is used as a frame buffer
                        add t2, #8              ' spin_gamepad = 7
                                                ' spin_tv_communication = 8 
                        ' unused here
                        
                        add t2, #4              ' spin_display_base = 9
                        mov display_base, t2    ' display base for scan buffer

                        mov fg_stop_scanline_ptr, horizontal_scroll
                        add fg_stop_scanline_ptr, #194<<1

                        mov tilemap_sub128, tilemap
                        sub tilemap_sub128, #128        ' optimisation: pre-subtract 128

                        ' prepare a 256 bytes buffer in HUB ram for decompression
                        ' place it just after the 102 bytes CLUT, thus overwriting tilemap and a part of foreground too
                        ' every cog (4) must have its own buffer
                        'mov hub_temp_buffer_ptr, tilemap
                        'add hub_temp_buffer_ptr, #256 

                        mov rle_pix_to_skip_first, #0
                        mov l1, cognumber wz
              if_z      jmp #:skip_loop_initial_skip
:loop_initial_skip      add rle_pix_to_skip_first, #256
                        'add hub_temp_buffer_ptr, #256
                        djnz l1, #:loop_initial_skip
:skip_loop_initial_skip
                        
                        ' Syncronise all gfx cog with TV driver before starting
                        call #waitsyncro_start
                        
main_loop
                        call #waitrequest_start

                        ' Good: all cogs will now prepare a scanline internally
                        rdword fg_stop_scanline, fg_stop_scanline_ptr

                        ' The current movie frame compressed data is stored in 'tiles' buffer
                        ' 'frame_ptr' is the pointer into RLE data
                        mov frame_ptr, tiles

                        ' indicate how many rle lines must be 'skipped'
                        ' Special case when starting the screen:
                        ' The number of lines to skip is equal to 'cognumber':
                        ' i.e.: cog 0 has NO line to skip, it must begin at top of picture
                        ' cog 1 must skip 1 line, etc.
                        ' Then afterward, every cog must skip exactly 'GFX_COG_TOTAL-1'*256 (i.e.: 3) lines in order
                        ' to reach their specific next line of data
                        mov rle_pix_to_skip, rle_pix_to_skip_first

                        ' shortcut special case only for cog#0 first line: pixel skip will be 0
                        cmp fg_stop_scanline, #511 wz
              if_e      jmp #blacktiles
                        tjz rle_pix_to_skip, #end_rle_skip_line
              
                        ' Here, prepare the scanline in internal memory
prepare_scanline        ' Special case: if stop_scanline is set to 511, it means: just display black for all scanlines
                        cmp fg_stop_scanline, #511 wz
              if_e      jmp #blacktiles

                        ' Loop here for 'rle_pix_to_skip' times
rle_skip_line           
                        
                        mov l1, rle_pix_to_skip            ' total number of pixel to emit
                        ' perform a 'silent rle decompression' without outputing anything on screen, just to skip for one line of data
rle_skip_line_loop_next
                        rdbyte rlelen, frame_ptr
rle_skip_line_loop_nex2 test rlelen, #128 wz
                        add frame_ptr, #1
                        rdbyte rlelen, frame_ptr
                        add rlelen, #1
               if_nz    djnz l1, #rle_skip_line_loop_nex2
               if_nz    jmp #end_rle_skip_line
rle_skip_continue
                        add frame_ptr, #1
                        sub l1, rlelen wz,wc
              if_a      jmp #rle_skip_line_loop_next
                        ' Warning: this optimisation saves 4 clocks, but makes the code very 'fragile'
                        ' if a corruption occurs, 'djnz' will possibly fall in an infinite loop
                        'sub l1, rlelen
                        'djnz l1, #rle_skip_line_loop_next

                        ' Here where are ready to emit one line of data 
end_rle_skip_line       mov rle_pix_to_skip, rle_pixel_skip_const 

                        
                        ' approx how many clocks per scanline do we have:
                        ' 80Mhz / 60 fps = 1333333 clocks per frame / 243 scanlines = 5486 clocks per line
                        ' minus TV copy (1272) and code overhead (100) = 4114 approx per line
                        ' 4114 / 256 pixels = 16 clocks per pixel
                        ' So we are running at 4 graphic cogs, we have approx: 16*4 = 64 clocks per pixel
                        ' Ouch! That's not a lot!                           

                        ' Method 1: unpack into internal cog scanbuffer,
                        ' then fast-copy to TV
'{
                        ' start a beginning of scanbuffer
                        movd copyfourpixel, #scanbuffer
                        ' total number of 4-pixel to emit:
                        mov l1, #64

                        ' prepare combine four-pixel output 
                        mov tilepixel, #0
                        ' 4-pixel counter
                        mov lshift, #4
loop_next_rle_batch                        
                        ' by default, set self-mod instruction to 'jmp'
                        mov four_pix_notdone_sm, four_pix_jmp_src
                        rdbyte rlecol, frame_ptr        ' read rle color (1 byte, 7 bit color + 1 bit for short rle)
                        test rlecol, #128 wz    ' if bit 128 is set, this is a rlelen=1 coded byte
                        ' Warning: this conditional jmp takes 8 clocks when skipped
              if_nz     jmp #:skip_rlelen_fetch
                        add frame_ptr, #1
                        rdbyte rlelen, frame_ptr ' else, we need to read the requested length (1 byte)
                        add rlecol, #128     ' get CLUT address
                        add rlelen, #1          ' add 1 to len since it is encoded as (length-1)
                        mov four_pix_notdone_sm, four_pix_notdone_src ' set 'djnz' instruction

:skip_rlelen_fetch
                        add rlecol, tilemap_sub128 ' subtract 128 and get CLUT address (optimized)
                        rdbyte rlecol, rlecol   ' perform color lookup for rlecol
                        add frame_ptr, #1
                        ' loop for 'rlelen' pixel and set them to 'rlecol'
loop_rlelen
                        or tilepixel, rlecol 
                        ror tilepixel, #8
                        djnz lshift, #four_pix_notdone_sm
                        ' good: a four-pixel batch is complete: emit it to the scanbuffer                        
copyfourpixel           mov scanbuffer, tilepixel  ' output 4 pixels on the screen
                        djnz l1, #continue
                        jmp #scanlinefinished
                        
continue                add copyfourpixel, destination_increment
                        mov tilepixel, #0
                        mov lshift, #4

' this instruction is self-modified depending on 'rlelen'
four_pix_notdone_sm     nop
four_pix_jmp_src        jmp #loop_next_rle_batch
' source inst for self-mod
four_pix_notdone_src    djnz rlelen, #loop_rlelen                        
'}

                        ' Method 2: unpack straight into HUB ram (taking advantage of wrbyte)
                        ' Then in TV_copy, transfer from this buffer to TV display buffer
{''
                        mov t3, hub_temp_buffer_ptr
                        ' total number of pixel to emit:
                        mov l1, #256

loop_next_rle_batch                        
                        ' by default, set self-mod instruction to 'jmp'
                        mov four_pix_notdone_sm, four_pix_jmp_src
                        rdbyte rlecol, frame_ptr        ' read rle color (1 byte, 7 bit color + 1 bit for short rle)
                        test rlecol, #128 wz    ' if bit 128 is set, this is a rlelen=1 coded byte
                        ' Warning: this conditional jmp takes 8 clocks when skipped
              if_nz     jmp #:skip_rlelen_fetch
                        add frame_ptr, #1
                        rdbyte rlelen, frame_ptr ' else, we need to read the requested length (1 byte)
                        add rlecol, #128     ' get CLUT address
                        add rlelen, #1          ' add 1 to len since it is encoded as (length-1)
                        mov four_pix_notdone_sm, four_pix_notdone_src ' set 'djnz' instruction

:skip_rlelen_fetch
                        add rlecol, tilemap_sub128 ' subtract 128 and get CLUT address (optimized)
                        rdbyte rlecol, rlecol   ' perform color lookup for rlecol
                        add frame_ptr, #1
                        ' loop for 'rlelen' pixel and set them to 'rlecol'
loop_rlelen             wrbyte rlecol, t3
                        add t3, #1 
                        djnz l1, #continue
                        jmp #scanlinefinished
continue                

' this instruction is self-modified depending on 'rlelen'
four_pix_notdone_sm     nop
four_pix_jmp_src        jmp #loop_next_rle_batch
' source inst for self-mod
four_pix_notdone_src    djnz rlelen, #loop_rlelen
''}

scanlinefinished        ' Check status of TV: warning: this is a pseudo-call: it might not return!
                        movd nextcopy1, #scanbuffer
                        movd nextcopy2, #scanbuffer+1
                        movd nextcopy3, #scanbuffer+2
                        movd nextcopy4, #scanbuffer+3
                        movd nextcopy5, #scanbuffer+4
                        movd nextcopy6, #scanbuffer+5
                        movd nextcopy7, #scanbuffer+6
                        movd nextcopy8, #scanbuffer+7
                        mov t1, display_base
                        mov l1, #64>>3

' Verification: if the TV is already asking for this scanline or more
' (except scanline 0), then lit the debug led: we have failed preparing this
' scanline in time: we need to optimize or use more cog
checktv                 'mov t3, hub_temp_buffer_ptr
                        'add t3, #252
                        {rdlong currentrequest, request_scanline
                        cmp currentrequest, currentscanline wz, wc
        if_ae           mov scanbuffer+63, white        ' emit a white 4-pixels at end of scanline to debug
        'if_ae           wrlong white, t3
        if_ae           jmp #start_tv_copy
                        }

:wait                   ' Wait here until the TV request exactly the scanline that THIS cog prepared
                        ' Other cog will wait here a bit
                        rdlong currentrequest, request_scanline
                        cmps currentrequest, currentscanline wz, wc
        if_b            jmp #:wait

start_tv_copy           ' Method 1: internal scanbuffer
                        ' approx: 78 clocks * 16 + 24 = 1272 clocks for TV copy
                        ' loop unrolled 4 times for slight speed increase
nextcopy1               wrlong scanbuffer, t1   'destination reg is self-modify
                        add nextcopy1, destination_increment8
                        add t1, #4
nextcopy2               wrlong scanbuffer, t1   'destination reg is self-modify
                        add nextcopy2, destination_increment8
                        add t1, #4
nextcopy3               wrlong scanbuffer, t1   'destination reg is self-modify
                        add nextcopy3, destination_increment8
                        add t1, #4
nextcopy4               wrlong scanbuffer, t1   'destination reg is self-modify
                        add nextcopy4, destination_increment8
                        add t1, #4
nextcopy5               wrlong scanbuffer, t1   'destination reg is self-modify
                        add nextcopy5, destination_increment8
                        add t1, #4
nextcopy6               wrlong scanbuffer, t1   'destination reg is self-modify
                        add nextcopy6, destination_increment8
                        add t1, #4
nextcopy7               wrlong scanbuffer, t1   'destination reg is self-modify
                        add nextcopy7, destination_increment8
                        add t1, #4
nextcopy8               wrlong scanbuffer, t1   'destination reg is self-modify
                        add nextcopy8, destination_increment8
                        add t1, #4 
                        
                        djnz l1, #nextcopy1

                        ' Method 2: copy from 'hub_temp_buffer_ptr' to TV:
                        {mov t3, hub_temp_buffer_ptr
                        mov t1, display_base
                        mov l1, #64>>1
                        
:nextcopy               rdlong t2, t3
                        add t3, #4
                        wrlong t2, t1
                        add t1, #4
                        wrlong t2, t1
                        add t1, #4
                        djnz l1, #:nextcopy }
                        
scanlinedone            
                        ' Line is done, increment to the next one this cog will handle
                        
add_cog_total           add currentscanline, #GFX_COG_TOTAL
                        cmp currentscanline, #191 wc, wz
        if_be           jmp #prepare_scanline
                        
                        ' The screen is completed, jump back to main loop a wait for next frame
                        jmp #main_loop

 
' optimised section of rle-skip
rle_skip_read_len       add frame_ptr, #1
                        djnz l1, #rle_skip_line_loop_next
                        jmp #end_rle_skip_line


' Output all black when drawing off-map stuff
blacktiles
                        call #debugclear
                        jmp #scanlinefinished

' Helper proc to clear the scanbuffer
debugclear
                        movd :debugclear, #scanbuffer
                        mov l1, #64
:debugclear             mov scanbuffer, #0
                        add :debugclear, destination_increment
                        djnz l1, #:debugclear
debugclear_ret          ret

' Wait until the tv driver request scanline 0
waitrequest_start
                        mov currentscanline, cognumber
:waitloop               rdlong currentrequest, request_scanline wz
        if_nz           jmp #:waitloop
waitrequest_start_ret   ret

' Wait until the tv driver request scanline 192
waitsyncro_start
:waitloop               rdlong currentrequest, request_scanline
                        cmp currentrequest, #192 wz
        if_ne           jmp #:waitloop
waitsyncro_start_ret    ret

                   

full long               $ffffffff
white long              $05050505               ' was $07070707 but now TV adds black
destination_increment long 512
destination_increment2 long 512<<1
destination_increment4 long 512<<2
destination_increment8 long 512<<3
zero long               $0
mask_pixel1 long        $000000FF
mask_pixel2 long        $0000FF00
mask_pixel3 long        $00FF0000
mask_pixel4 long        $FF000000
rle_pixel_skip_const long     (GFX_COG_TOTAL-1)*256

end_of_cog_ram          long end_of_cog_ram

cognumber               res     1
request_scanline        res
vertical_scroll         res
horizontal_scroll       res
sprite_adr              res
tilemap                 res
tiles                   res
display_base            res

currentscanline         res     1
currentrequest          res     1
tilepixel               res     1
lshift                  res     1
rshift                  res     1
fg_stop_scanline_ptr    res     1
fg_stop_scanline        res     1

frame_ptr               res
rlecol                  res
rlelen                  res
rle_pix_to_skip         res
rle_pix_to_skip_first   res
tilemap_sub128          res
hub_temp_buffer_ptr     res

' t1-8: temporary registers
t1                      res     1
t2                      res     1
t3                      res     1
t4                      res     1
t5                      res     1
' l1-5: loop registers
l1                      res     1
l2                      res     1
l3                      res     1
l4                      res     1
l5                      res     1
' p1-7: parameter registers
p1                      res     1
p2                      res     1


fit 429
scanbuffer res          64
