''*************************************
''*  Rem Graphic engine v020          *
''*************************************

CON

  GFX_COG_TOTAL = 4
  
  MAX_NB_SPRITE = 24   


  ' Reminder: the screen is 256x192, which gives 32x24 tiles.
  ' Limitation:
  ' MAPSIZEX must be a power of 2.
  ' MAPSIZEY have no restriction, can be from 1 to infinity.
  'NOTE these 3 const must by reflected in 'rem_drv_tv' 
  MAPSIZEX = 32
  MAPSIZEY = 32
  MAPSIZEX_SHIFT = 5    ' MAPSIZEX_SHIFT must be set according with MAPSIZEX

  'Map dimension for FOREGROUND map
  FGSIZEX = 32
  FGSIZEY = 32
  FGSIZEX_SHIFT = 5    ' FGSIZEX_SHIFT must be set according with FGSIZEX
  

{PUB start(cognum, paramadr)

'' Start REM engine - starts a cog
  coginit(cognum, @entry, paramadr)
}
pub RV_getcog_overlap
  return @entry
  
PUB RV_getcog_end_of_ram
  return (end_of_cog_ram-1)<<2


DAT

                        org
' Entry
'
entry                   'mov dira, #1 ' enable debug led port
                        ' fetch some parameter, starting at 'par':
                        rdlong cognumber, par ' store cog number (0,1,...)
                        mov t2, #255            ' signal that we acknowledged our cognumber
                        wrlong t2, par           

                        ' Wait here until 'par' points to 511 (magic value to make gfx cogs starts)
wait_startup            rdlong t2, par
                        cmp t2, #511 wz
              if_nz     jmp #wait_startup

                        mov t2, par             ' prefect some pointer to speed up runtime execution:
                        add t2, #4              '   spin_request_scanline = 1
                        mov request_scanline, t2
                        add t2, #4              ' spin_vertical_scroll = 2      
                        rdlong vertical_scroll, t2
                        add t2, #4              ' spin_horizontal_scroll = 3          
                        rdlong horizontal_scroll, t2
                        add t2, #4              ' spin_sprites_adr = 4     
                        rdlong sprite_adr, t2
                        add t2, #4              ' spin_tilemap_adr = 5        
                        rdlong tilemap, t2
                        add t2, #4              ' spin_tiles_adr = 6    
                        rdlong tiles, t2
                        add t2, #8              ' spin_gamepad = 7
                                                ' spin_tv_communication = 8 
                        ' unused here
                        
                        add t2, #4              ' spin_display_base = 9
                        mov display_base, t2    ' display base for scan buffer

                        ' Emit black into last 8 pixels, which are used to hide tile/scrolling popping on the right side of screen
                        mov t1, display_base
                        add t1, #248
                        wrlong framecount, t1     
                        add t1, #4
                        wrlong framecount, t1      

                        mov fg_tilemap, #MAPSIZEY
                        shl fg_tilemap, #MAPSIZEX_SHIFT
                        add fg_tilemap, tilemap ' calculate foreground tilemap pointer, starting just after the bg map 

                        ' Prepare vertical_base_offset and horizontal_base_offset pointers
                        mov vertical_base_offset, vertical_scroll
                        add vertical_base_offset, #192<<1
                        mov horizontal_base_offset, horizontal_scroll
                        add horizontal_base_offset, #192<<1
                        mov vertical_fg_offset, vertical_scroll
                        add vertical_fg_offset, #193<<1
                        mov horizontal_fg_offset, horizontal_scroll
                        add horizontal_fg_offset, #193<<1
                        mov fg_start_scanline_ptr, vertical_scroll
                        add fg_start_scanline_ptr, #194<<1
                        mov fg_stop_scanline_ptr, horizontal_scroll
                        add fg_stop_scanline_ptr, #194<<1

                        ' Syncronise all gfx cog with TV driver before starting
                        call #waitsyncro_start
                        
main_loop
                        call #waitrequest_start

                        ' Special case: cog #0 will prepare all sprite animation address offset
        '                cmp cognumber, #0 wz
        'if_z            call #setspriteanim

'                        add framecount, #1
        
                        ' Good: all cogs will now prepare a scanline internally
                        rdword fg_start_scanline, fg_start_scanline_ptr
                        rdword fg_stop_scanline, fg_stop_scanline_ptr

                        ' Here, prepare the scanline in internal memory
prepare_scanline

                        ' Special case: if stop_scanline is set to 511, it means: just display black for all scanlines
                        cmp fg_stop_scanline, #511 wz
        if_e            jmp #blacktiles

                        ' Take the currentscanline and add the vertical_scroll
                        ' new method with vertical scroll per scanline
                        mov y_offset, currentscanline
                        shl y_offset, #1        ' mult by 2 since it is word spin_y_offsets[192]
                        add y_offset, vertical_scroll
                        rdword y_offset, y_offset

                        rdword t1, vertical_base_offset  ' add base offset
                        add y_offset, t1 

                        ' new method: horizontal scroll per scanline:
                        mov x_offset, currentscanline
                        shl x_offset, #1        ' mult by 2 since it is word spin_x_offsets[192]
                        add x_offset, horizontal_scroll
                        rdword x_offset, x_offset

                        rdword t1, horizontal_base_offset  ' add base offset
                        add x_offset, t1
                        shr x_offset, #6        ' fixed-point: >>6 for BG x_offset 

                        rdword fg_x_offset, horizontal_fg_offset  ' FG: only have a base offset
                        shr fg_x_offset, #4        ' fixed-point: >>4 for x_offset
                        rdword fg_y_offset, vertical_fg_offset  ' FG: only have a base offset
                        add fg_y_offset, currentscanline
                        sub fg_y_offset, fg_start_scanline ' auto-offset on 'start scanline'

                        
                        'mov sprite_a_offset, x_offset  ' enable this for sprite horizontal scrolling

                        ' Check if currentscanline has reached 'stop_y': this is where we stop
                        ' generating any new pixel data. This will be used to perform a 'curtain screen'
                        ' wipe effect.
                        {rdlong stop_y, stop_y_adr
                        cmp currentscanline, stop_y wc, wz
        if_ae           jmp #scanlinedone

                        ' check if we are in the 'top y portion' of the screen (i.e.: non-scrolling)
                        rdlong top_y, top_y_adr
                        cmps currentscanline, top_y wc, wz
        if_b            jmp #top_portion

                        rdlong bottom_y, bottom_y_adr
                        cmps currentscanline, bottom_y wc, wz
        if_ae           jmp #bottom_portion}

                        ' safety check: if trying to draw a line off the tilemap, then output a black scanline
        '                cmp y_offset, max_tile_y wc, wz
        'if_a            jmp #blacktiles

                        ' Start at tilemap 0,0
                        ' tilemap_offset = tilemap offset counter
                        ' skip 32 first tiles because they represent the top and bottom non-scrolling portion
                        'mov tilemap_offset, #32
                        ' deprecated
                        ' add (currentscanline / 8) * MAPSIZEX
                        mov t1, y_offset
                        shr t1, #3
                        shl t1, #MAPSIZEX_SHIFT
                        mov tilemap_offset, t1

                        movd copyfourpixel, #scanbuffer_prev
                        movd copyfourpixel2, #scanbuffer
                        mov l1, #33

                        ' Rewind by 4 pixels if x_offset is above 4
                        mov t1, x_offset
                        and t1, #7
                        cmp t1, #4              wz,wc
              if_ae     sub copyfourpixel, destination_increment      
              if_ae     sub copyfourpixel2, destination_increment

                        ' Calculate things that won't change during a scanline:
                        mov lshift, x_offset
                        and lshift, #3
                        mov t1, lshift wz
        if_nz           mov shift_rpixel, shift_rpixel_instruction ' set self-modify code
        if_nz           mov shift_rpixel2, shift_rpixel_instruction ' set self-modify code
        if_z            mov shift_rpixel, skip_rpixel_instruction ' to either shift or move
        if_z            mov shift_rpixel2, skip_rpixel_instruction ' to either shift or move
                        ' I have to do this because shr and shl can't shift by 32 bit :(
                        shl lshift, #3

                        mov rshift, #4
                        sub rshift, t1
                        shl rshift, #3
                        
                        add tilemap_offset, tilemap

                        mov y_tile_offset, y_offset
                        and y_tile_offset, #7
                        shl y_tile_offset, #3
                        add y_tile_offset, tiles

                        ' Optimized 'background layer' rendering:
                        ' each tile is 8 pixel wide,
                        ' so this loop will render 8 pixels (2 longs)
fetchtile
                        mov t2, tilemap_offset

                        ' Now add x offset / 8 : this will give us a full tile offset
                        mov t1, x_offset
                        shr t1, #3
                        and t1, #(MAPSIZEX-1)  ' Horizontal wrap: this only works if map width is a power of 2
                        add t2, t1
                        
                        ' Fetch the current tile in t2
                        rdbyte t2, t2
                        shl t2, #6              ' Each tile is 64 bytes
                        ' t2 will point to the tile memory
                         '  add the current scanline & 7 line offset for this tile and the combined 32-bit x-offset
                        add t2, y_tile_offset

                        rdlong t1, t2           ' 4 Pixels read into t1
                        mov tilepixel, t1
shift_rpixel            shl tilepixel, rshift
                        or tilepixel, nexttilepixel
                        
                        mov nexttilepixel, t1
                        shr nexttilepixel, lshift
                        
copyfourpixel           mov scanbuffer, tilepixel  ' output 4 pixels on the screen
                        add copyfourpixel, destination_increment2

                        add t2, #4              ' read the other 4 pixel of this same tile
                        rdlong t1, t2           ' 4 Pixels read into t1
                        mov tilepixel, t1
shift_rpixel2           shl tilepixel, rshift
                        or tilepixel, nexttilepixel
                        
                        mov nexttilepixel, t1
                        shr nexttilepixel, lshift
                        
copyfourpixel2          mov scanbuffer, tilepixel  ' output 4 pixels on the screen
                        add copyfourpixel2, destination_increment2
                        {
                        ' with JMP, DJNZ, CALL, MOV, RDBYTE conditional: '3' flick
                        mov t1, #20 wz                        
:test_speed  if_always       rdbyte t2,t2
:next_inst             djnz t1, #:test_speed 
                        }
                        ' advance to next eight pixel
                        add x_offset, #8
                        djnz l1, #fetchtile


bg_map_finished
                        ' check if scanline is less than 'start_scanline'
                        cmp currentscanline, fg_start_scanline wc, wz
        if_b            jmp #scanlinefinished
                        ' check if scanline is greater 'stop_scanline'
                        cmp currentscanline, fg_stop_scanline wc, wz
        if_a            jmp #scanlinefinished

                        ' Now render the foreground tilemap over with transparency
                        mov t1, fg_y_offset
                        shr t1, #3
                        shl t1, #FGSIZEX_SHIFT
                        mov fg_tilemap_offset, t1

                        ' reset pointer to start of scanbuffer
                        movd fg_copyfourpixel, #scanbuffer_prev ' ..but go back 32-bit because we'll output 65 long
                        movs fg_andfourpixel, #scanbuffer_prev ' ..but go back 32-bit because we'll output 65 long
                        movd fg_copyfourpixel2, #scanbuffer
                        movs fg_andfourpixel2, #scanbuffer

                        ' Rewind by 4 pixels if x_offset is above 4
                        mov t1, fg_x_offset
                        and t1, #7
                        cmp t1, #4              wz,wc
              if_ae     sub fg_copyfourpixel, destination_increment      
              if_ae     sub fg_andfourpixel, #1      
              if_ae     sub fg_copyfourpixel2, destination_increment
              if_ae     sub fg_andfourpixel2, #1

                        ' number of eight-pixel to output = 32 (since the last one is hidden)
                        mov l1, #32
                        
                        ' Calculate things that won't change during a scanline:
                        mov lshift, fg_x_offset
                        and lshift, #3
                        mov t1, lshift wz
        if_nz           mov fg_shift_rpixel, shift_rpixel_instruction ' set self-modify code
        if_nz           mov fg_shift_rpixel2, shift_rpixel_instruction ' set self-modify code
        if_z            mov fg_shift_rpixel, skip_rpixel_instruction ' to either shift or move
        if_z            mov fg_shift_rpixel2, skip_rpixel_instruction ' to either shift or move
                        ' I have to do this because shr and shl can't shift by 32 bit :(
                        shl lshift, #3

                        mov rshift, #4
                        sub rshift, t1
                        shl rshift, #3

                        add fg_tilemap_offset, fg_tilemap

                        mov fg_y_tile_offset, fg_y_offset
                        and fg_y_tile_offset, #7
                        shl fg_y_tile_offset, #3
                        add fg_y_tile_offset, tiles

                        mov full, full wc       ' set carry
                        mov t5, #0              ' t5 required for Z->C transfer 

' Fetch tile (32-bit) pixel into 'tilepixel'
fg_fetchtile
                        mov t2, fg_tilemap_offset

                        ' Now add x offset / 8 : this will give us a full tile offset
                        mov t1, fg_x_offset
                        shr t1, #3
                        and t1, #(FGSIZEX-1)  ' Horizontal wrap: this only works if map width is a power of 2
                        add t2, t1
                        
                        ' Fetch the current tile in t2
                        rdbyte t2, t2           
                        shl t2, #6              ' Each tile is 64 bytes
                        ' t2 will point to the tile memory
                         '  add the current scanline & 7 line offset for this tile
                        ' finally add the 32-bit x-offset
                        add t2, fg_y_tile_offset

                        ' Technique 3: edge management using 'Z' and 'C' flags
                        ' We determine if a tile is fully transparent or fully opaque depending on its previous neighbor
                        ' speed: best case=8-10 sprites, worst=6 sprites (worst is 1/0/1/0/... tile which isn't realistic)
                        rdlong t1, t2 wz           ' Pixel read into t1
                        mov tilepixel, t1
fg_shift_rpixel         shl tilepixel, rshift
                        or tilepixel, nexttilepixel
                        
                        mov nexttilepixel, t1
                        shr nexttilepixel, lshift

                        ' Perform edge decision:
                        ' Z flag = 1 if the tile is transparent, =0 if opaque
                        ' C flag = 1 if last tile was transparent, like a sequence of transparent tile
                        ' C flag = 0 when drawing a sequence of opaque block 
        if_nz_and_nc    jmp #fg_copyfourpixel   'If Z=0 and C=0, this 4-pixel is 100% opaque, no need to mask
        IF_Z_AND_C      jmp #fg_skip_tile       ' Fully transparent: We can skip this 8-pixel completely if Z=1 and C=1

                        muxz t5, #1 wc          ' copy Z to C

fg_andfourpixel         mov t3, scanbuffer
                        test tilepixel, mask_pixel1 wz
        if_nz           andn t3, mask_pixel1
                        test tilepixel, mask_pixel2 wz
        if_nz           andn t3, mask_pixel2
                        test tilepixel, mask_pixel3 wz
        if_nz           andn t3, mask_pixel3
                        test tilepixel, mask_pixel4 wz
        if_nz           andn t3, mask_pixel4
                        
                        or tilepixel, t3 
                        
fg_copyfourpixel        mov scanbuffer, tilepixel

                        ' second batch of 4 pix:
                        add t2, #4              ' read the other 4 pixel of this same tile
                        rdlong t1, t2 'wz           ' Pixel read into t1
                        mov tilepixel, t1
fg_shift_rpixel2        shl tilepixel, rshift
                        or tilepixel, nexttilepixel
                        
                        mov nexttilepixel, t1
                        shr nexttilepixel, lshift

                        ' Perform edge decision:
                        ' Z flag = 1 if the tile is transparent, =0 if opaque
                        ' C flag = 1 if last tile was transparent, like a sequence of transparent tile
                        ' C flag = 0 when drawing a sequence of opaque block 
        if_nc           jmp #fg_copyfourpixel2   'If Z=0 and C=0, this 4-pixel is 100% opaque, no need to mask
        'IF_Z_AND_C      jmp #fg_skip_tile2       ' We can skip this 4-pixel completely if Z=1 and C=1

                        'muxz t5, #1 wc          ' copy Z to C

fg_andfourpixel2        mov t3, scanbuffer
                        test tilepixel, mask_pixel1 wz
        if_nz           andn t3, mask_pixel1
                        test tilepixel, mask_pixel2 wz
        if_nz           andn t3, mask_pixel2
                        test tilepixel, mask_pixel3 wz
        if_nz           andn t3, mask_pixel3
                        test tilepixel, mask_pixel4 wz
        if_nz           andn t3, mask_pixel4
                        
                        or tilepixel, t3
fg_copyfourpixel2       mov scanbuffer, tilepixel

fg_skip_tile            add fg_andfourpixel, #2
                        add fg_copyfourpixel, destination_increment2
                        add fg_andfourpixel2, #2
                        add fg_copyfourpixel2, destination_increment2
                        
                        ' advance to next four pixel
                        add fg_x_offset, #8
                        djnz l1, #fg_fetchtile


fg_map_finished
                        ' Here we draw the sprites!
                        ' Between each sprite, we should do a checktv in case we spent too much time in this line

startsprite
                        ' Draw shadow sprite. Disabled for this game.
                        ' If you want to enable shadow, just uncomment this chunk of code
                        {
                        mov t1, framecount
                        and t1, #1 wz
        if_nz           jmp #skip_shadow

                        mov spriteoutput, shadowoutput_instruction
                        mov master_sprite_offset_x, #4
                        mov master_sprite_offset_y, #6
                        call #drawspriteproc
                        }
skip_shadow
                        mov spriteoutput, spriteoutput_instruction
                        'mov master_sprite_offset_x, #0
                        'mov master_sprite_offset_y, #0
                        call #drawspriteproc
                        ' Mask out the last 8 pixels since we are streaming map data directly into the last visible column on screen
                        ' (or else the new tiles will 'pop' at the right side of the screen)
                        'mov scanbuffer+62, #0
                        'mov scanbuffer+63, #0
                        jmp #scanlinefinished

drawspriteproc
                        mov l5, #MAX_NB_SPRITE   ' TODO: l5 is 'number of sprites'
                        
                        ' Let's iterate for each sprites
                        ' The first sprite starts at firstsprite_adr
                        mov sprite_x_adr, sprite_adr
                        mov sprite_y_adr, sprite_adr
                        add sprite_y_adr, #2
                        mov sprite_s_adr, sprite_adr
                        add sprite_s_adr, #4
                        mov sprite_p_adr, sprite_adr
                        add sprite_p_adr, #6

nextsprite
                        ' If address == 0, this sprite is invisible / disabled.
                        rdword sprite_p, sprite_p_adr wz
        if_z            jmp #spritedone

                        ' Perform Y check:
                        ' sprite_y = currentscanline - sprite_y
                        ' if sprite_y < 0, out
                        ' if sprite_y >= height, out
                        ' Fetch sprite_h and sprite_wshift by unpacking from sprite_s_adr
                        rdword sprite_h, sprite_s_adr
                        mov sprite_wshift, sprite_h
                        shr sprite_wshift, #8
                        and sprite_h, #$ff

                        rdword t1, sprite_y_adr
                        shl t1, #16             ' perform manual 'sign-extend'
                        sar t1, #16+7           ' X and Y pos are fixed-point 7-bit
                        mov sprite_y, currentscanline
                        sub sprite_y, t1
                        'sub sprite_y, master_sprite_offset_y
                        
                        cmp sprite_y, sprite_h wc, wz
        if_ae           jmp #spritedone

                        ' sprite_s contains the 'shift width': translate this into a real pixel width here
                        mov sprite_s, #1
                        shl sprite_s, sprite_wshift

                        ' Perform X check:
                        rdword sprite_x, sprite_x_adr
                        shl sprite_x, #16          ' perform manual 'sign-extend'
                        sar sprite_x, #16+7        ' X and Y pos are fixed-point 7-bit
                        ' Adjust x according to horizontal map scroll
                        'sub sprite_x, sprite_a_offset
                        'add sprite_x, master_sprite_offset_x
                        
                        mov t1, sprite_x
                        add t1, sprite_s
                        cmps t1, #0 wc, wz
        if_be           jmp #spritedone
                        
                        cmps sprite_x, #256 wc, wz
        if_ae           jmp #spritedone

                        mov l1, sprite_s
                        shr l1, #2
                        add l1, #1

                        ' Check if sprite is partially outside screen on the left
                        sub t1, #256 wc, wz
        if_a            shr t1, #2              ' Yes? So let's skip group of four pixel that are outside
        if_a            sub l1, t1
                        
                        ' Check if sprite is partially outside screen on the right
                        mov t1, sprite_x
                        cmps t1, #0 wc, wz
        if_ae           jmp #:x_not_negative
                        ' Sprite is partially outside screen
                        neg t1, t1
                        shr t1, #2

                        ' substract number of pixel to output because some of them are outside
                        sub l1, t1

                        ' skip some pixel data
                        shl t1, #2
                        add sprite_p, t1

                        add sprite_x, t1
        
:x_not_negative         sub sprite_x, #1
                        mov lshift, sprite_x

                        ' Set starting X position in scanbuffer
                        shr sprite_x, #2
                        mov t1, #scanbuffer
                        add t1, sprite_x
                        movd spriteoutput, t1
                        movd spriteoutputand, t1
                        '
                        mov t1, sprite_y
                        shl t1, sprite_wshift
                        add sprite_p, t1
                        
                        and lshift, #3
                        neg lshift, lshift
                        add lshift, #3
                        
                        mov t1, lshift wz
        if_nz           mov shift_rpixel_sprite, shift_rpixel_instruction ' set self-modify code
        if_z            mov shift_rpixel_sprite, skip_rpixel_instruction ' to either shift or move
                        ' I have to do this because shr and shl can't shift by 32 bit :(
                        shl lshift, #3

                        mov rshift, #4
                        sub rshift, t1
                        shl rshift, #3

                        mov nexttilepixel, #0
                        
nextspritepixel         
                        cmp l1, #1 wz           ' Nullify last pixel
        if_e            mov t1, #0
        if_ne           rdlong t1, sprite_p ' Pixel read into t1
        
                        mov tilepixel, t1
shift_rpixel_sprite     shl tilepixel, rshift
                        or tilepixel, nexttilepixel wz

                        mov nexttilepixel, t1
                        shr nexttilepixel, lshift

        if_z            jmp #skipblackpixel     ' Optimisation: skip this whole 32-bit pack if it's all 0s.

                        ' here tilepixel contains the four pixel we want to output
                        ' we'll build a mask in t3 for all non-0 pixel
                        mov t3, full

                        test tilepixel, mask_pixel1 wz
        if_nz           xor t3, mask_pixel1
                        test tilepixel, mask_pixel2 wz
        if_nz           xor t3, mask_pixel2
                        test tilepixel, mask_pixel3 wz
        if_nz           xor t3, mask_pixel3
                        test tilepixel, mask_pixel4 wz
        if_nz           xor t3, mask_pixel4

spriteoutputand         and scanbuffer, t3
spriteoutput            or scanbuffer, tilepixel        ' TODO: Replace this instruction by nop for shadow sprite
skipblackpixel
                        add spriteoutputand, destination_increment
                        add spriteoutput, destination_increment
                        add sprite_p, #4
                        ' advance to next four pixel
                        djnz l1, #nextspritepixel

spritedone
                        call #checktv           ' Check if we have enough time for another sprite
                        
                        add sprite_p_adr, #8
                        add sprite_x_adr, #8
                        add sprite_y_adr, #8
                        add sprite_s_adr, #8

                        djnz l5, #nextsprite
drawspriteproc_ret      ret

scanlinefinished        ' Check status of TV: warning: this is a pseudo-call: it might not return!
                        'call #checktv           

:wait                   ' Wait here until the TV request exactly the scanline that THIS cog prepared
                        ' Other cog will wait here a bit
                        rdlong currentrequest, request_scanline
                        cmps currentrequest, currentscanline wz, wc
        if_b            jmp #:wait

start_tv_copy
                        movd :nextcopy, #scanbuffer
                        movd :nextcopy2, #scanbuffer+1
                        mov t1, display_base
                        mov l1, #62>>1             ' only 62*4 = 248 pixels are being copied: we don't emit the last tile used to hide scrolling/tile popping

:nextcopy               wrlong scanbuffer, t1   'destination reg is self-modify
                        add :nextcopy, destination_increment2
                        add t1, #4
:nextcopy2              wrlong scanbuffer, t1   'destination reg is self-modify
                        add :nextcopy2, destination_increment2
                        add t1, #4
                        djnz l1, #:nextcopy
                        
scanlinedone            
                        ' Line is done, increment to the next one this cog will handle
                        
add_cog_total           add currentscanline, #GFX_COG_TOTAL
                        cmp currentscanline, #191 wc, wz
        if_be           jmp #prepare_scanline
                        
                        ' The screen is completed, jump back to main loop a wait for next frame
                        jmp #main_loop

' Instruction that will get copied over at line 'shift_rpixel'
shift_rpixel_instruction shl tilepixel, rshift
skip_rpixel_instruction mov tilepixel, #0
 
spriteoutput_instruction or scanbuffer, tilepixel
shadowoutput_instruction nop

' Verification: if the TV is already asking for this scanline or more
' (except scanline 0), then lit the debug led: we have failed preparing this
' scanline in time: we need to optimize or use more cog
' if we failed, then skip straight to TV output: This means that WE WONT RETURN TO THE CALLER.
checktv
                        tjz currentrequest, #checktv_ret
                        rdlong currentrequest, request_scanline
                        cmp currentrequest, currentscanline wz, wc
                        ' At this point, we could skip drawing sprites or something like that and go
                        ' straight to image output and still stay synced with the tv driver
                        ' We must be quick though!
        'if_b            mov outa, #0
        'if_ae           mov outa, #1
        if_ae           jmp #start_tv_copy
checktv_ret             ret
        

' Output all black when drawing off-map stuff
blacktiles
                        movd :debugclear, #scanbuffer
                        mov l1, #64
:debugclear             mov scanbuffer, #0
                        add :debugclear, destination_increment
                        djnz l1, #:debugclear
                        
                        jmp #scanlinefinished

' Wait until the tv driver request scanline 0
waitrequest_start
                        mov currentscanline, cognumber
:waitloop               rdlong currentrequest, request_scanline wz
        if_nz           jmp #:waitloop
waitrequest_start_ret   ret

' Wait until the tv driver request scanline 192
waitsyncro_start
:waitloop               rdlong currentrequest, request_scanline
                        cmp currentrequest, #192 wz
        if_ne           jmp #:waitloop
waitsyncro_start_ret    ret

' Complete used memory for gfx driver code: about 376 longs (1504 bytes)                        

full long               $ffffffff
white long              $05050505               ' was $07070707 but now TV adds black
destination_increment long 512
destination_increment2 long 512<<1
framecount long         $0
mask_pixel1 long        $000000FF
mask_pixel2 long        $0000FF00
mask_pixel3 long        $00FF0000
mask_pixel4 long        $FF000000
max_tile_y long         (MAPSIZEY*16-1)

end_of_cog_ram          long end_of_cog_ram

cognumber               res     1
request_scanline        res
vertical_scroll         res
horizontal_scroll       res
sprite_adr              res
tilemap                 res
tiles                   res
display_base            res

sprite_x_adr            res     1
sprite_y_adr            res     1
sprite_s_adr            res     1
sprite_p_adr            res     1

currentscanline         res     1
currentrequest          res     1
tilemap_offset          res     1
tilepixel               res     1
nexttilepixel           res     1
lshift                  res     1
rshift                  res     1
y_offset                res     1
x_offset                res     1
y_tile_offset           res     1
'top_y                   res     1
'bottom_y                res     1
'stop_y                  res     1
'sprite_a_offset         res     1              ' Note: enable lines in code if you need sprites to follow horizontal map scroll offset
vertical_base_offset    res     1
horizontal_base_offset  res     1
fg_tilemap              res     1 
vertical_fg_offset      res     1
horizontal_fg_offset    res     1
fg_y_offset             res     1
fg_x_offset             res     1
fg_y_tile_offset        res     1
fg_tilemap_offset       res     1
fg_start_scanline_ptr   res     1
fg_stop_scanline_ptr    res     1
fg_start_scanline       res     1
fg_stop_scanline        res     1

sprite_x                res     1
sprite_y                res     1
sprite_s                res     1
sprite_p                res     1
sprite_h                res     1
sprite_wshift           res     1

'master_sprite_offset_x  res     1
'master_sprite_offset_y  res     1

' t1-8: temporary registers
t1                      res     1
t2                      res     1
t3                      res     1
t4                      res     1
t5                      res     1
' l1-5: loop registers
l1                      res     1
l2                      res     1
l3                      res     1
l4                      res     1
l5                      res     1
' p1-7: parameter registers
p1                      res     1
p2                      res     1


fit 429
scanbuffer_prev res     1                       ' One extra long on the left
scanbuffer res          66                      ' two extra long on the right